# -*- coding: utf-8 -*-
"""This module defines functions for fetching and parsing files in the PDB  
for the Biologically Interesting Molecule Reference Dictionary (BIRD_). 

.. _BIRD: https://www.wwpdb.org/data/bird

The chemical information is stored in Peptide Reference Dictionary (PRD) files, 
whereas the biological function is documented in a separate family file."""

import os.path

from prody import LOGGER, getPackagePath
from prody.utilities import isListLike
from prody.proteins import wwPDBServer, WWPDB_FTP_SERVERS, parseSTAR

__all__ = ['fetchBIRDviaFTP', 'parseBIRD']

def fetchBIRDviaFTP(**kwargs):
    """Retrieve the whole Biologically Interesting Molecule Reference 
    Dictionary (BIRD) resource, which is updated every week. This includes 
    2 kinds of keys, which can be selected with the **keys** keyword argument.

    The chemical information is found in a zipped (tar.gz) directory at 
    https://files.rcsb.org/pub/pdb/data/bird/prd/prd-all.cif.gz, which 
    contains individual CIF files within it. This data will be downloaded 
    and extracted to :file:`.prody/bird-prd`.

    Biological function information is also found in a zipped (tar.gz) directory at 
    https://files.rcsb.org/pub/pdb/data/bird/family/family-all.cif.gz, which 
    contains individual CIF files within it. This data will be downloaded 
    and extracted to :file:`.prody/bird-family`.

    :arg keys: keys specifying which data to fetch out of ``'prd'``, ``'family'`` or ``'both'``
               default is ``'both'``
    :type keys: str, tuple, list, :class:`~numpy.ndarray`

    The underlying data can be accessed using :func:`parseBIRD`."""

    BIRD_PATH = os.path.join(getPackagePath(), 'bird')

    keys = kwargs.get('keys', 'both')
    if isinstance(keys, str):
        if keys == 'both':
            keys = ['prd', 'family']
        elif keys[:3].lower() == 'prd':
            keys = ['prd']
        elif keys[:3].lower() == 'fam':
            keys = ['family']
        else:
            raise ValueError("keys should be 'both', 'prd' or 'fam'")

    elif isListLike(keys):
        keys = list(keys)
    else:
        raise TypeError("keys should be list-like or string")

    ftp_divided = 'pdb/data/bird/'
    ftp_pdbext = '.cif.gz'
    ftp_prefix = ''

    if not os.path.isdir(BIRD_PATH):
        os.mkdir(BIRD_PATH)

    LOGGER.progress('Downloading BIRD', len(keys),
                    '_prody_fetchBIRD')

    ftp_name, ftp_host, ftp_path = WWPDB_FTP_SERVERS[wwPDBServer() or 'us']
    LOGGER.debug('Connecting wwPDB FTP server {0}.'.format(ftp_name))

    from ftplib import FTP
    try:
        ftp = FTP(ftp_host)
    except Exception as error:
        raise type(error)('FTP connection problem, potential reason: '
                          'no internet connectivity')
    else:
        count = 0
        success = 0
        failure = 0
        filenames = []
        ftp.login('')
        for i, x in enumerate(keys):
            data = []
            ftp_fn = ftp_prefix + '{0}-all'.format(x) + ftp_pdbext
            try:
                ftp.cwd(ftp_path)
                ftp.cwd(ftp_divided)
                ftp.cwd(x)
                ftp.retrbinary('RETR ' + ftp_fn, data.append)
            except Exception as error:
                if ftp_fn in ftp.nlst():
                    LOGGER.warn('{0} download failed ({1}). It is '
                                'possible that you do not have rights to '
                                'download .gz files in the current network.'
                                .format(x, str(error)))
                else:
                    LOGGER.info('{0} download failed. {1} does not exist '
                                'on {2}.'.format(ftp_fn, x, ftp_host))
                failure += 1
                filenames.append(None)
            else:
                if len(data):
                    filename = BIRD_PATH + '/{0}-all.cif.gz'.format(x)

                    with open(filename, 'w+b') as outfile:
                        write = outfile.write
                        [write(block) for block in data]

                    success += 1
                else:
                    failure += 1
            count += 1
            LOGGER.update(i, label='_prody_fetchBIRD')
        LOGGER.finish()

    LOGGER.debug('PDB download via FTP completed ({0} downloaded, '
                 '{1} failed).'.format(success, failure))


def parseBIRD(*ids, **kwargs):
    """Parse data from the Biologically Interesting Molecule Reference 
    Dictionary (BIRD) resource, which is updated every week. This includes 
    2 kinds of keys, which can be selected with the **keys** keyword argument.

    The chemical information is found in a single CIF file at 
    https://files.rcsb.org/pub/pdb/data/bird/prd/prd-all.cif.gz. 
    This data will be downloaded and extracted to :file:`.prody/bird-prd`.

    Biological function information is also found in a single CIF file at 
    https://files.rcsb.org/pub/pdb/data/bird/family/family-all.cif.gz. 
    This data will be downloaded and extracted to :file:`.prody/bird-family`.

    Individual compounds can be selected using **ids**. 
    If needed, BIRD files are downloaded using :func:`.fetchBIRDviaFTP` function.
    
    You can also provide arguments that you would like passed on to fetchBIRDviaFTP.

    :arg ids: one BIRD identifier (starting with PRD or FAM) or a list of them.
        If **None** is provided then all of them are returned.
    :type ids: str, tuple, list, :class:`~numpy.ndarray`, **None**

    :arg key: key specifying which data to fetch out of ``'prd'`` or ``'family'``
               default is ``'prd'``
    :type key: str

    Returns :class:`.StarDataBlock` object or list of them.
    """
    key = kwargs.get('key', 'prd')
    if not isinstance(key, str):
        raise TypeError("key should be a string")

    if key[:3].lower() == 'prd':
        key = 'prd'
    elif key[:3].lower() == 'fam':
        key = 'family'
    else:
        raise ValueError("key should be 'prd' or 'fam'")

    n_ids = len(ids)
    if n_ids == 1:
        if isListLike(ids[0]):
            ids = ids[0]
            n_ids = len(ids)

    if n_ids == 1:
        ids = list(ids)

    BIRD_PATH = os.path.join(getPackagePath(), 'bird')
    filename = BIRD_PATH + '/{0}-all.cif.gz'.format(key)
    if not os.path.isfile(filename):
        fetchBIRDviaFTP(keys=key, **kwargs)

    data = parseSTAR(filename, shlex=True)
    ret = []
    for id in ids:
        try:
            ret.append(data[list(data.search(id)._dict.keys())[0]])
        except ValueError:
            try:
                ret.append(data[id])
            except ValueError:
                LOGGER.warn('id {0} not found in {1} data '
                            'so appending None'.format(id, key))
                ret.append(None)

    if n_ids == 1:
        return ret[0]

    return ret



"""This module defines miscellaneous utility functions."""

from textwrap import wrap

__all__ = ['joinLinks', 'joinRepr', 'joinTerms', 'tabulate', 'wrapText']


def joinLinks(links, sep=', ', last=None, sort=False):
    """Returns a string joining *links* as reStructuredText."""

    links = list(links)
    if sort:
        links.sort()
    if last is None:
        last = ''
    else:
        last = last + '`' + links.pop() + '`_'
    return '`' + ('`_' + sep + '`').join(links) + '`_' + last


def joinRepr(items, sep=', ', last=None, sort=False):
    """Returns a string joining representations of *items*."""

    items = [repr(item) for item in items]
    if sort:
        items.sort()
    if last is None:
        last = ''
    else:
        last = last + items.pop()
    return sep.join(items) + last


def joinTerms(terms, sep=', ', last=None, sort=False):
    """Returns a string joining *terms* as reStructuredText."""

    terms = list(terms)
    if sort:
        terms.sort()
    if last is None:
        last = ''
    else:
        last = last + ':term:`' + terms.pop() + '`'
    return ':term:`' + ('`' + sep + ':term:`').join(terms) + '`' + last


def wrapText(text, width=70, join='\n', **kwargs):
    """Returns wrapped lines from :func:`textwrap.wrap` after *join*ing them.
    """

    try:
        indent = kwargs.pop('indent')
    except KeyError:
        pass
    else:
        kwargs['initial_indent'] = kwargs['subsequent_indent'] = ' ' * indent
    if join:
        return join.join(wrap(text, width, **kwargs))
    else:
        return wrap(text, width, **kwargs)

def tabulate(*cols, **kwargs):
    """Returns a table for columns of data.

    :kwarg header: make first row a header, default is **True**
    :type header: bool
    :kwarg width: 79
    :type width: int
    :kwargs space: number of white space characters between columns,
         default is 2
    :type space: int

    """

    indent = kwargs.get('indent', 0)
    space = kwargs.get('space', 2)
    widths = [max(map(len, cols[0]))] # PY3K: OK
    widths.append(kwargs.get('width', 79) - sum(widths) -
                  len(widths) * space)
    space *= ' '
    bars = (space).join(['=' * width for width in widths])
    lines = [bars]

    for irow, items in enumerate(zip(*cols)): # PY3K: OK
        rows = []
        rows.extend([wrap(item, widths[icol]) if icol else
                      [item.ljust(widths[icol])]
                         for icol, item in enumerate(items)])
        maxlen = max(map(len, rows)) # PY3K: OK
        if maxlen > 1:
            for i, row in enumerate(rows):
                row.extend([' ' * widths[i]] * (maxlen - len(row)))
        for line in zip(*rows):
            lines.append(space.join(line))
        if not irow and kwargs.get('header', True):
            lines.append(bars)
    if irow > 1:
        lines.append(bars)
    return '\n'.join(lines)



# -*- coding: utf-8 -*-
"""This module defines :class:`Angle` for dealing with angle information provided
by using :meth:`.AtomGroup.setAngles` method."""

from numbers import Integral
import numpy as np

RAD2DEG = 180 / np.pi

__all__ = ['Angle']

class Angle(object):

    """A pointer class for angled atoms.  Following built-in functions are
    customized for this class:

    * :func:`len` returns angle length, i.e. :meth:`getSize`
    * :func:`iter` yields :class:`~.Atom` instances"""

    __slots__ = ['_ag', '_acsi', '_indices']

    def __init__(self, ag, indices, acsi=None):

        self._ag = ag
        self._indices = np.array(indices)
        if acsi is None:
            self._acsi = ag.getACSIndex()
        else:
            self._acsi = acsi

    def __repr__(self):

        one, two, three = self._indices
        names = self._ag._getNames()
        return '<Angle: {0}({1})--{2}({3})--{4}({5}) from {6}>'.format(
                            names[one], one, names[two], two, 
                            names[three], three, str(self._ag))

    def __str__(self):

        one, two, three = self._indices
        names = self._ag._getNames()
        return '{0}({1})--{2}({3})--{4}({5})'.format(
                                            names[one], one, names[two], two,
                                            names[three], three)

    def __eq__(self, other):

        return (isinstance(other, Angle) and other.getAtomGroup() is self._ag
                and (np.all(other.getIndices() == self._indices) or
                 np.all(other.getIndices() == list(reversed(self._indices)))))

    def __ne__(self, other):

        return not self.__eq__(other)

    def __size__(self):

        return self.getSize()

    def __iter__(self):

        for index in self._indices:
            yield self._ag[index]

    def getAtomGroup(self):
        """Returns atom group."""

        return self._ag

    def getAtoms(self):
        """Returns angled atoms."""

        return (self._ag[self._indices[0]], self._ag[self._indices[1]], self._ag[self._indices[2]])

    def getIndices(self):
        """Returns indices of angled atoms."""

        return self._indices.copy()

    def getSize(self, radian=False):
        """Returns angle size."""

        v1, v2 = self.getVectors()
        rad = np.arccos((v1*v2).sum(-1) / ((v1**2).sum(-1) * (v2**2).sum(-1))**0.5)
        if radian:
            return rad
        else:
            return rad * RAD2DEG

    def getVectors(self):
        """Returns bond vectors that originate from the central atom."""

        one, two, three = self._indices
        acsi = self.getACSIndex()
        vector1 = self._ag._coords[acsi, one] - self._ag._coords[acsi, two]
        vector2 = self._ag._coords[acsi, three] - self._ag._coords[acsi, two]
        return vector1, vector2

    def getACSIndex(self):
        """Returns index of the coordinate set."""

        acsi = self._acsi
        if acsi >= self._ag._n_csets:
            raise ValueError('{0} has fewer coordsets than assumed by {1}'
                             .format(str(self._ag), str(self)))
        return acsi

    def setACSIndex(self, index):
        """Set the coordinate set at *index* active."""

        if self._ag._coords is None:
            raise AttributeError('coordinates are not set')

        if not isinstance(index, Integral):
            raise TypeError('index must be an integer')

        n_csets = self._ag._n_csets
        if n_csets <= index or n_csets < abs(index):
            raise IndexError('coordinate set index is out of range')

        if index < 0:
            index += n_csets

        self._acsi = index


def evalAngles(angles, n_atoms):
    """Returns an array mapping atoms to their angled neighbors and an array
    that stores number of angles made by each atom."""

    numangles = np.bincount(angles.reshape((angles.shape[0] * 3)))
    angmap = np.zeros((n_atoms, numangles.max(), 2), int)
    angmap.fill(-1)
    index = np.zeros(n_atoms, int)
    for angle in angles:
        a, b, c = angle
        angmap[a, index[a]] = [b,c]
        angmap[b, index[b]] = [a,c]
        angmap[c, index[c]] = [a,b]
        index[angle] += 1
    return angmap, numangles


def trimAngles(angles, indices):
    """Returns angles between atoms at given indices."""

    iset = set(indices)
    angles = [angle for angle in angles if angle[0] in iset and angle[1] in iset and angle[2] in iset]
    if angles:
        newindices = np.zeros(indices.max()+1, int)
        newindices[indices] = np.arange(len(indices))
        return newindices[np.array(angles)]



"""Merge multiple MSAs based on common labels."""

__author__ = 'Ahmet Bakan, Anindita Dutta'

from ..apptools import DevelApp

__all__ = ['evol_merge']

APP = DevelApp('merge', 'merge multiple MSAs based on common labels')

APP.setExample(
"""Merges multiple MSAs into one MSA based on common labels
appearing across all the input MSAs. The following example shows
how to merge two MSA of a multi-domain protein.

  $ evol search 3KG2A
  $ evol fetch ANF_receptor
  $ evol fetch Lig_chan
  $ evol merge ANF_receptor_full.slx Lig_chan_full.slx
""", [])


APP.addArgument('msa',
    nargs='+',
    help='MSA filenames to be merged')

APP.addGroup('output', 'output options')
APP.addArgument('-o', '--outname',
    dest='outname',
    help='output filename, default is first input filename '
         'with _merged suffix',
    type=str,
    metavar='STR',
    group='output')

APP.addArgument('-f', '--format',
    dest='format',
    type=str,
    metavar='STR',
    help='output MSA file format, default is same as first input MSA',
    group='output')

APP.addArgument('-z', '--compressed',
    dest='compressed',
    action='store_true',
    help='gzip merged MSA output',
    group='output')

def evol_merge(*msa, **kwargs):

    import prody
    from prody import parseMSA, mergeMSA, LOGGER, writeMSA, MSAFile
    from prody.sequence.msafile import MSAEXTMAP
    from os.path import splitext
    if len(msa) < 2:
        raise ValueError('multiple msa filenames must be specified')
    msaobj = []
    try:
        msaobj = [parseMSA(fn) for fn in msa]
    except:
        raise IOError('failed to parse {0}'.format(fn))

    msafile = MSAFile(msa[0])

    format = kwargs.get('format') or msafile.format
    outname = kwargs.get('outname') or (msafile.getTitle() + '_merged' +
                                        MSAEXTMAP[msafile.format])
    writeMSA(outname, mergeMSA(*msaobj), **kwargs)
    LOGGER.info('Merged MSA is saved as: {0}'.format(outname))

APP.setFunction(evol_merge)




"""Refine MSA application."""

from ..apptools import DevelApp

__all__ = ['evol_refine']

APP = DevelApp('refine', 'refine an MSA by removing gapped rows/colums')

APP.setExample(
"""Refines MSA by removing gapped columns (residue positions)
and rows (sequences).

In the following, columns that are gaps in sequence with label
GTHB2_ONCKE will be removed from MSA:

  $ evol refine piwi.slx -l GTHB2_ONCKE""", [])


APP.addArgument('msa',
    help='MSA filename to be refined')

APP.addGroup('refine', 'refinement options')
APP.addArgument('-l', '--label',
    dest='label',
    help='sequence label, UniProt ID code, or PDB and chain identifier',
    type=str,
    metavar='STR',
    group='refine')

APP.addArgument('-s', '--seqid',
    dest='seqid',
    help='identity threshold for selecting unique sequences',
    default=None,
    type=float,
    metavar='FLOAT',
    group='refine')

APP.addArgument('-c', '--colocc',
    dest='colocc',
    help='column (residue position) occupancy',
    default=None,
    type=float,
    metavar='FLOAT',
    group='refine')

APP.addArgument('-r', '--rowocc',
    dest='rowocc',
    help='row (sequence) occupancy',
    default=None,
    type=float,
    metavar='FLOAT',
    group='refine')

APP.addArgument('-k', '--keep',
    dest='pdbres',
    help='keep columns corresponding to residues not resolved in PDB '
         'structure, applies label argument is a PDB identifier',
    default=False,
    action='store_true',
    group='refine')

APP.addGroup('output', 'output options')
APP.addArgument('-o', '--outname',
    dest='outname',
    help='output filename, default is msa filename with _refined suffix',
    type=str,
    metavar='STR',
    group='output')

APP.addArgument('-f', '--format',
    dest='format',
    type=str,
    metavar='STR',
    help='output MSA file format, default is same as input',
    group='output')

APP.addArgument('-z', '--compressed',
    dest='compressed',
    action='store_true',
    help='gzip refined MSA output',
    group='output')

def evol_refine(msa, **kwargs):

    from prody import parseMSA, refineMSA, writeMSA, LOGGER
    from os.path import splitext

    outname = kwargs.get('outname')
    if outname is None:
        outname, ext = splitext(msa)
        if ext.lower() == '.gz':
            outname, _ = splitext(msa)
        outname += '_refined' + ext

    writeMSA(outname, refineMSA(parseMSA(msa), **kwargs), **kwargs)
    LOGGER.info('Refined MSA is written in file: ' + outname)


APP.setFunction(evol_refine)



# -*- coding: utf-8 -*-
"""This module defines classes for principal component analysis (PCA) and
essential dynamics analysis (EDA) calculations."""

import time

import numpy as np

from prody import LOGGER, PY2K
from prody.atomic import Atomic
from prody.ensemble import Ensemble, PDBEnsemble
from prody.trajectory import TrajBase
from prody.utilities import importLA, solveEig, ZERO

from .nma import NMA

if PY2K:
    range = xrange

__all__ = ['PCA', 'EDA']


class PCA(NMA):

    """A class for Principal Component Analysis (PCA) of conformational
    ensembles. See examples in :ref:`pca`."""

    def __init__(self, name='Unknown'):

        NMA.__init__(self, name)

    def setCovariance(self, covariance, is3d=True):
        """Set covariance matrix."""

        if not isinstance(covariance, np.ndarray):
            raise TypeError('covariance must be an ndarray')
        elif not (covariance.ndim == 2 and
                  covariance.shape[0] == covariance.shape[1]):
            raise ValueError('covariance must be square matrix')
        elif covariance.dtype != float:
            try:
                covariance = covariance.astype(float)
            except:
                raise ValueError('covariance.dtype must be float')

        self._reset()

        self._is3d = is3d
        self._cov = covariance
        self._dof = covariance.shape[0]

        if is3d:
            self._n_atoms = self._dof // 3
        else:
            self._n_atoms = self._dof
        self._trace = self._cov.trace()

    def buildCovariance(self, coordsets, **kwargs):
        """Build a covariance matrix for *coordsets* using mean coordinates
        as the reference.  *coordsets* argument may be one of the following:

        * :class:`.Atomic`
        * :class:`.Ensemble`
        * :class:`.TrajBase`
        * :class:`numpy.ndarray` with shape ``(n_csets, n_atoms, 3)``

        For ensemble and trajectory objects, ``update_coords=True`` argument
        can be used to set the mean coordinates as the coordinates of the
        object.

        When *coordsets* is a trajectory object, such as :class:`.DCDFile`,
        covariance will be built by superposing frames onto the reference
        coordinate set (see :meth:`.Frame.superpose`).  If frames are already
        aligned, use ``aligned=True`` argument to skip this step.


        .. note::
           If *coordsets* is a :class:`.PDBEnsemble` instance, coordinates are
           treated specially.  Let's say **C**\_ij is the element of the
           covariance matrix that corresponds to atoms *i* and *j*.  This
           super element is divided by number of coordinate sets (PDB models or
           structures) in which both of these atoms are observed together."""

        if not isinstance(coordsets, (Ensemble, Atomic, TrajBase, np.ndarray)):
            raise TypeError('coordsets must be an Ensemble, Atomic, Numpy '
                            'array instance')
        LOGGER.timeit('_prody_pca')
        mean = None
        weights = None
        ensemble = None
        if isinstance(coordsets, np.ndarray):
            if (coordsets.ndim != 3 or coordsets.shape[2] != 3 or
                    coordsets.dtype not in (np.float32, float)):
                raise ValueError('coordsets is not a valid coordinate array')
        elif isinstance(coordsets, Atomic):
            coordsets = coordsets._getCoordsets()
        elif isinstance(coordsets, Ensemble):
            ensemble = coordsets
            if isinstance(coordsets, PDBEnsemble):
                weights = coordsets.getWeights() > 0
            coordsets = coordsets._getCoordsets()

        update_coords = bool(kwargs.get('update_coords', False))
        quiet = kwargs.pop('quiet', False)

        if isinstance(coordsets, TrajBase):
            nfi = coordsets.nextIndex()
            coordsets.reset()
            n_atoms = coordsets.numSelected()
            dof = n_atoms * 3
            cov = np.zeros((dof, dof))
            #mean = coordsets._getCoords().flatten()
            n_confs = 0
            n_frames = len(coordsets)
            if not quiet:
                LOGGER.info('Covariance will be calculated using {0} frames.'
                            .format(n_frames))
            coordsum = np.zeros(dof)
            if not quiet:
                LOGGER.progress('Building covariance', n_frames, '_prody_pca')
            align = not kwargs.get('aligned', False)
            for frame in coordsets:
                if align:
                    frame.superpose()
                coords = frame._getCoords().flatten()
                coordsum += coords
                cov += np.outer(coords, coords)
                n_confs += 1
                if not quiet:
                    LOGGER.update(n_confs, label='_prody_pca')
            if not quiet:
                LOGGER.finish()
            cov /= n_confs
            coordsum /= n_confs
            mean = coordsum
            cov -= np.outer(coordsum, coordsum)
            coordsets.goto(nfi)
            self._cov = cov
            if update_coords:
                coordsets.setCoords(mean.reshape((n_atoms, 3)))
        else:
            n_confs = coordsets.shape[0]
            if n_confs < 3:
                raise ValueError('coordsets must have more than 3 coordinate '
                                 'sets')
            n_atoms = coordsets.shape[1]
            if n_atoms < 3:
                raise ValueError('coordsets must have more than 3 atoms')
            dof = n_atoms * 3
            if not quiet:
                LOGGER.info('Covariance is calculated using {0} coordinate sets.'
                            .format(len(coordsets)))
            s = (n_confs, dof)
            if weights is None:
                if coordsets.dtype == float:
                    self._cov = np.cov(coordsets.reshape((n_confs, dof)).T,
                                       bias=1)
                else:
                    cov = np.zeros((dof, dof))
                    coordsets = coordsets.reshape((n_confs, dof))
                    mean = coordsets.mean(0)
                    if not quiet:
                        LOGGER.progress('Building covariance', n_confs,
                                    '_prody_pca')
                    for i, coords in enumerate(coordsets.reshape(s)):
                        deviations = coords - mean
                        cov += np.outer(deviations, deviations)
                        if not quiet:
                            LOGGER.update(n_confs, label='_prody_pca')
                    if not quiet:
                        LOGGER.finish()
                    cov /= n_confs
                    self._cov = cov
            else:
                # PDB ensemble case
                mean = np.zeros((n_atoms, 3))
                for i, coords in enumerate(coordsets):
                    mean += coords * weights[i]
                mean /= weights.sum(0)
                d_xyz = ((coordsets - mean) * weights).reshape(s)
                divide_by = weights.astype(float).repeat(3, axis=2).reshape(s)
                self._cov = np.dot(d_xyz.T, d_xyz) / np.dot(divide_by.T,
                                                            divide_by)
            if update_coords and ensemble is not None:
                if mean is None:
                    mean = coordsets.mean(0)
                ensemble.setCoords(mean)

        self._trace = self._cov.trace()
        self._dof = dof
        self._n_atoms = n_atoms
        if not quiet:
            LOGGER.report('Covariance matrix calculated in %2fs.', '_prody_pca')

    def calcModes(self, n_modes=20, turbo=True, **kwargs):
        """Calculate principal (or essential) modes.  This method uses
        :func:`scipy.linalg.eigh`, or :func:`numpy.linalg.eigh`, function
        to diagonalize the covariance matrix.

        :arg n_modes: number of non-zero eigenvalues/vectors to calculate,
            default is 20,
            if **None** or ``'all'`` is given, all modes will be calculated
        :type n_modes: int

        :arg turbo: when available, use a memory intensive but faster way to
            calculate modes, default is **True**
        :type turbo: bool"""
        
        if self._cov is None:
            raise ValueError('covariance matrix is not built or set')
        start = time.time()
        self._clear()
        if str(n_modes).lower() == 'all':
            n_modes = None
        
        values, vectors, _ = solveEig(self._cov, n_modes=n_modes, zeros=True, 
                                      turbo=turbo, reverse=True, **kwargs)
        which = values > ZERO
        self._eigvals = values[which]
        self._array = vectors[:, which]
        self._vars = values[which]
        self._n_modes = len(self._eigvals)
        if self._n_modes > 1:
            LOGGER.debug('{0} modes were calculated in {1:.2f}s.'
                     .format(self._n_modes, time.time()-start))
        else:
            LOGGER.debug('{0} mode was calculated in {1:.2f}s.'
                     .format(self._n_modes, time.time()-start))

    def performSVD(self, coordsets):
        """Calculate principal modes using singular value decomposition (SVD).
        *coordsets* argument may be a :class:`.Atomic`, :class:`.Ensemble`,
        or :class:`numpy.ndarray` instance.  If *coordsets* is a numpy array,
        its shape must be ``(n_csets, n_atoms, 3)``.  Note that coordinate
        sets must be aligned prior to SVD calculations.

        This is a considerably faster way of performing PCA calculations
        compared to eigenvalue decomposition of covariance matrix, but is
        an approximate method when heterogeneous datasets are analyzed.
        Covariance method should be preferred over this one for analysis of
        ensembles with missing atomic data.  See :ref:`pca-xray-calculations`
        example for comparison of results from SVD and covariance methods."""

        linalg = importLA()

        start = time.time()
        if not isinstance(coordsets, (Ensemble, Atomic, np.ndarray)):
            raise TypeError('coordsets must be an Ensemble, Atomic, Numpy '
                            'array instance')
        if isinstance(coordsets, np.ndarray):
            if (coordsets.ndim != 3 or coordsets.shape[2] != 3 or
                    coordsets.dtype not in (np.float32, float)):
                raise ValueError('coordsets is not a valid coordinate array')
            deviations = coordsets - coordsets.mean(0)
        else:
            if isinstance(coordsets, Ensemble):
                deviations = coordsets.getDeviations()
            elif isinstance(coordsets, Atomic):
                deviations = (coordsets._getCoordsets() -
                              coordsets._getCoords())

        n_confs = deviations.shape[0]
        if n_confs <= 3:
            raise ValueError('coordsets must have more than 3 coordinate sets')
        n_atoms = deviations.shape[1]
        if n_atoms <= 3:
            raise ValueError('coordsets must have more than 3 atoms')

        dof = n_atoms * 3
        deviations = deviations.reshape((n_confs, dof)).T

        vectors, values, self._temp = linalg.svd(deviations,
                                                 full_matrices=False)
        values = (values ** 2) / n_confs
        self._dof = dof
        self._n_atoms = n_atoms
        which = values > 1e-18
        self._eigvals = values[which]
        self._array = vectors[:, which]
        self._vars = self._eigvals
        self._trace = self._vars.sum()
        self._n_modes = len(self._eigvals)
        LOGGER.debug('{0} modes were calculated in {1:.2f}s.'
                     .format(self._n_modes, time.time()-start))

    def addEigenpair(self, eigenvector, eigenvalue=None):
        """Add eigen *vector* and eigen *value* pair(s) to the instance.
        If eigen *value* is omitted, it will be set to 1.  Eigenvalues
        are set as variances."""

        NMA.addEigenpair(self, eigenvector, eigenvalue)
        self._vars = self._eigvals

    def setEigens(self, vectors, values=None):
        """Set eigen *vectors* and eigen *values*.  If eigen *values* are
        omitted, they will be set to 1.  Eigenvalues are set as variances."""

        self._clear()
        NMA.setEigens(self, vectors, values)
        self._vars = self._eigvals


class EDA(PCA):

    """A class for Essential Dynamics Analysis (EDA) [AA93]_.
    See examples in :ref:`eda`.

    .. [AA93] Amadei A, Linssen AB, Berendsen HJ. Essential dynamics of
       proteins. *Proteins* **1993** 17(4):412-25."""

    pass



import numpy as np

from prody.dynamics import NMA, MaskedGNM
from prody.dynamics.mode import Mode
from prody.dynamics.modeset import ModeSet

from prody.utilities import importLA, copy, showFigure, div0
from prody import LOGGER, SETTINGS

__all__ = ['showDomains', 'showEmbedding', 'getDomainList']

## normalization methods ##

def showDomains(domains, linespec='-', **kwargs):
    """A convenient function that can be used to visualize Hi-C structural domains. 
    *kwargs* will be passed to :func:`matplotlib.pyplot.plot`.

    :arg domains: a 2D array of Hi-C domains, such as [[start1, end1], [start2, end2], ...].
    :type domains: :class:`numpy.ndarray`
    """

    fill_ends = kwargs.pop('fill_ends', 'close')
    domains = np.array(domains)
    shape = domains.shape

    if len(shape) == 1:
        # convert to domain list if labels are provided
        indicators = np.diff(domains)
        length = len(domains)
        if fill_ends in ['open', 'close']:
            indicators = np.append(1., indicators)
            indicators[-1] = 1
        elif fill_ends == 'skip':
            indicators = np.append(0., indicators)
        else:
            raise ValueError('invalid fill_ends mode: %s'%str(fill_ends))
        sites = np.where(indicators != 0)[0]
        starts = sites[:-1]
        ends = sites[1:]
        domains = np.array([starts, ends]).T
        consecutive = True
    elif len(shape) == 2:
        if domains.dtype == bool:
            length = domains.shape[1]
            domains_ = []
            for h in domains:
                start = None
                for i, b in enumerate(h):
                    if b:
                        if start is None:  # start
                            start = i
                    else:
                        if start is not None: # end
                            domains_.append([start, i-1])
                            start = None
                if start is not None:
                    domains_.append([start, i])
            domains = np.array(domains_)
        else:
            length = domains.max()
        consecutive = False
    else:
        raise ValueError('domains must be either one or two dimensions')

    from matplotlib.pyplot import figure, plot

    a = []; b = []
    lwd = kwargs.pop('linewidth', 1)
    lwd = kwargs.pop('lw', lwd)
    linewidth = np.abs(lwd)
    if fill_ends == 'open' and len(domains) == 1:
        domains = []

    for i in range(len(domains)):
        domain = domains[i]
        start = domain[0]; end = domain[1]
        if fill_ends == 'open' and start == 0:
            a.extend([end, end])
            b.extend([start, end])
        elif fill_ends == 'open' and end == length-1:
            a.extend([start, end])
            b.extend([start, start])
        else:
            a.extend([start, end, end])
            b.extend([start, start, end])

        if not consecutive:
            a.append(np.nan)
            b.append(np.nan)

    if lwd > 0:
        x = a; y = b
    else:
        x = b; y = a

    plt = plot(x, y, linespec, linewidth=linewidth, **kwargs)
    if SETTINGS['auto_show']:
        showFigure()
    return plt

def _getEigvecs(modes, row_norm=False, dummy_mode=False):
    la = importLA()

    if isinstance(modes, (Mode, ModeSet, NMA)):
        if hasattr(modes, '_model'):
            model = modes._model
        else:
            model = modes
        if isinstance(model, MaskedGNM):
            masked = model.masked
            model.masked = True
            V = modes.getArray()
            model.masked = masked
        else:
            V = modes.getArray()
    elif isinstance(modes, np.ndarray):
        V = modes
    else:
        try:
            mode0 = modes[0]
            if isinstance(mode0, Mode):
                V = np.empty((len(mode0),0))
                for mode in modes:
                    assert isinstance(mode, Mode), 'Modes should be a list of modes.'
                    v = mode.getEigvec()
                    v = np.expand_dims(v, axis=1)
                    V = np.hstack((V, v))
            else:
                V = np.array(modes)
        except TypeError:
            raise TypeError('Modes should be a list of modes.')
    if V.ndim == 1:
        V = np.expand_dims(V, axis=1)

    # add a dummy zero mode to the modeset
    if dummy_mode:
        v0 = V[:, 0]
        if np.allclose(v0, np.mean(v0)):
            dummy_mode = False
            LOGGER.warn('at least one zero mode is detected therefore dummy mode will NOT be added')

    if dummy_mode:
        n, _ = V.shape
        v0 = np.ones((n, 1), dtype=V.dtype)
        v0 /= la.norm(v0)
        V = np.hstack((v0, V))
        LOGGER.debug('a dummy zero mode is added')

    # normalize the rows so that feature vectors are unit vectors
    if row_norm:
        norms = la.norm(V, axis=1)
        N = np.diag(div0(1., norms))
        V = np.dot(N, V)

    return V

def showEmbedding(modes, labels=None, trace=True, headtail=True, cmap='prism'):
    """Visualizes Laplacian embedding of Hi-C data. 

    :arg modes: modes in which loci are embedded. It can only have 2 or 3 modes for the purpose 
    of visualization.
    :type modes: :class:`.ModeSet`

    :arg labels: a list of integers indicating the segmentation of the sequence.
    :type labels: list

    :arg trace: if **True** then the trace of the sequence will be indicated by a grey dashed line.
    :type trace: bool

    :arg headtail: if **True** then a star and a closed circle will indicate the head and the tail 
    of the sequence respectively.
    :type headtail: bool

    :arg cmap: the color map used to render the *labels*.
    :type cmap: str
    """
    V, _ = _getEigvecs(modes, True)
    m, n = V.shape

    if labels is not None:
        if len(labels) != m:
            raise ValueError('Modes (%d) and the Hi-C map (%d) should have the same number'
                                ' of atoms. Turn off "masked" if you intended to apply the'
                                ' modes to the full map.'
                                %(m, len(labels)))
    if n > 3:
        raise ValueError('This function can only visualize the embedding of 2 or 3 modes.')
    
    from matplotlib.pyplot import figure, plot, scatter
    from mpl_toolkits.mplot3d import Axes3D

    if n == 2:
        la = importLA()

        X, Y = V[:,:2].T
        R = np.array(range(len(X)))
        R = R / la.norm(R)
        X *= R; Y *= R
        
        f = figure()
        if trace:
            plot(X, Y, ':', color=[0.3, 0.3, 0.3])
        if labels is None:
            C = 'b'
        else:
            C = labels
        scatter(X, Y, s=30, c=C, cmap=cmap)
        if headtail:
            plot(X[:1], Y[:1], 'k*', markersize=12)
            plot(X[-1:], Y[-1:], 'ko', markersize=12)
    elif n == 3:
        X, Y, Z = V[:,:3].T
        
        f = figure()
        ax = f.add_subplot(1,1,1,projection="3d")
        if trace:
            ax.plot(X, Y, Z, ':', color=[0.3, 0.3, 0.3])
        if labels is None:
            C = 'b'
        else:
            C = labels
        ax.scatter(X, Y, Z, s=30, c=C, depthshade=True, cmap=cmap)
        if headtail:
            ax.plot(X[:1], Y[:1], Z[:1], 'k*', markersize=12)
            ax.plot(X[-1:], Y[-1:], Z[-1:], 'ko', markersize=12)

    if SETTINGS['auto_show']:
        showFigure()
    return f

def getDomainList(labels):
    """Returns a list of domain separations. The list has two columns: the first is for 
    the domain starts and the second is for the domain ends."""

    indicators = np.diff(labels)
    indicators = np.append(1., indicators)
    indicators[-1] = 1
    sites = np.where(indicators != 0)[0]
    starts = sites[:-1]
    ends = sites[1:]
    domains = np.array([starts, ends]).T

    return domains



from numbers import Integral

from numpy import ma
import numpy as np
from scipy.sparse import coo_matrix
from scipy.stats import mode
from prody.chromatin.norm import VCnorm, SQRTVCnorm, Filenorm
from prody.chromatin.functions import div0, showDomains, _getEigvecs

from prody import PY2K
from prody.dynamics import GNM, MaskedGNM
from prody.dynamics.functions import writeArray
from prody.dynamics.mode import Mode
from prody.dynamics.modeset import ModeSet

from prody.utilities import openFile, importLA, showMatrix, isURL, fixArraySize, makeSymmetric

__all__ = ['HiC', 'parseHiC', 'parseHiCStream', 'parseHiCBinary', 'saveHiC', 'loadHiC', 'writeMap']

class HiC(object):

    """This class is used to store and preprocess Hi-C contact map. A :class:`.GNM`
    instance for analyzing the contact map can be also created by using this class.
    """

    def __init__(self, title='Unknown', map=None, bin=None):
        self._title = title
        self._map = None
        self.mask = False
        self._labels = 0
        self.masked = True
        self.bin = bin
        self.map = map
    
    @property
    def map(self):
        if self.masked:
            return self.getTrimedMap()
        else:
            return self._map

    @map.setter
    def map(self, value):
        if value is None: 
            self._map = None
        else:
            self._map = np.asarray(value)
            self._map = makeSymmetric(self._map)
            self._maskUnmappedRegions()
            self._labels = np.zeros(len(self._map), dtype=int)

    def __repr__(self):
        mask = self.mask
        
        if np.isscalar(mask):
            return '<HiC: {0} ({1} loci)>'.format(self._title, len(self._map))
        else:
            return '<HiC: {0} ({1} mapped loci; {2} in total)>'.format(self._title, np.count_nonzero(mask), len(self._map))

    def __str__(self):

        return 'HiC ' + self._title

    def __getitem__(self, index):
        if isinstance(index, Integral):
            return self.map.flatten()[index]
        else:
            i, j = index
            return self.map[i,j]

    def __len__(self):
        mask = self.mask 
        
        if np.isscalar(mask):
            return len(self._map)
        else:
            return np.count_nonzero(mask)
    
    def numAtoms(self):
        return len(self.map)

    def getTitle(self):
        """Returns title of the instance."""

        return self._title

    def setTitle(self, title):
        """Sets title of the instance."""

        self._title = str(title)

    def getCompleteMap(self):
        """Obtains the complete contact map with unmapped regions."""

        return self._map
        
    def getTrimedMap(self):
        """Obtains the contact map without unmapped regions."""

        if self._map is None: 
            return None
        if np.isscalar(self.mask):
            return self._map

        M = ma.array(self._map)
        M.mask = np.diag(~self.mask)
        return ma.compress_rowcols(M)
    
    def align(self, array, axis=None):
        if not isinstance(array, np.ndarray):
            array = np.array(array)

        ret = array = array.copy()

        if np.isscalar(self.mask):
            return ret

        mask = self.mask.copy()

        l_full = self.getCompleteMap().shape[0]
        l_trim = self.getTrimedMap().shape[0]
        
        if len(array.shape) == 0:
            raise ValueError('array cannot be empty')
        elif len(array.shape) == 1:
            l = array.shape[0]
            if l == l_trim:
                N = len(mask)
                ret = np.zeros(N, dtype=array.dtype)
                ret[mask] = array
            elif l == l_full:
                ret = array[mask]
            else:
                raise ValueError('The length of array (%d) does not '
                                'match that of either the full (%d) '
                                'or trimed (%d).'
                                %(l, l_full, l_trim))
        elif len(array.shape) == 2:
            s = array.shape

            if axis is None:
                if s[0] != s[1]:
                    raise ValueError('The array must be a square matrix '
                                     'if axis is set to None.')
                if s[0] == l_trim:
                    N = len(mask)
                    whole_mat = np.zeros((N,N), dtype=array.dtype)
                    mask = np.outer(mask, mask)
                    whole_mat[mask] = array.flatten()
                    ret = whole_mat
                elif s[0] == l_full:
                    M = ma.array(array)
                    M.mask = np.diag(mask)
                    ret = ma.compress_rowcols(M)
                else:
                    raise ValueError('The size of array (%d) does not '
                                    'match that of either the full (%d) '
                                    'or trimed (%d).'
                                    %(s[0], l_full, l_trim))
            else:
                new_shape = list(s)
                otheraxis = 0 if axis!=0 else 1
                if s[axis] == l_trim:
                    N = len(mask)
                    new_shape[axis] = N
                    whole_mat = np.zeros(new_shape)
                    mask = np.expand_dims(mask, axis=otheraxis)
                    mask = mask.repeat(s[otheraxis], axis=otheraxis)
                    whole_mat[mask] = array.flatten()
                    ret = whole_mat
                elif s[axis] == l_full:
                    mask = np.expand_dims(mask, axis=otheraxis)
                    mask = mask.repeat(s[otheraxis])
                    ret = self._map[mask]
                else:
                    raise ValueError('The size of array (%d) does not '
                                    'match that of either the full (%d) '
                                    'or trimed (%d).'
                                    %(s[0], l_full, l_trim))
        
        return ret

    def getKirchhoff(self):
        """Builds a Kirchhoff matrix based on the contact map."""

        if self._map is None:
            return None
        else:
            M = self.map
            
            I = np.eye(M.shape[0], dtype=bool)
            A = M.copy()
            A[I] = 0.
            D = np.diag(np.sum(A, axis=0))
            K = D - A
            return K

    def _maskUnmappedRegions(self, diag=False):
        """Finds and masks unmapped regions in the contact map."""

        M = self._map
        if M is None: return

        if diag:
            # Obtain the diagonal values, need to make sure d is an array 
            # instead of a matrix, otherwise diag() later will not work as 
            # intended.
            d = np.array(np.diag(M))
        else:
            d = np.array(M.sum(0))

        # mask if a diagonal value is zero
        mask_zero = np.array(d==0)
        # mask if a diagonal value is NAN
        mask_nan = np.isnan(d)
        # combine two masks
        mask = np.logical_or(mask_nan, mask_zero)
        self.mask = ~mask

        return self.mask
    
    def calcGNM(self, n_modes=None, **kwargs):
        """Calculates GNM on the current Hi-C map. By default, ``n_modes`` is 
        set to **None** and ``zeros`` to **True**."""

        if 'zeros' not in kwargs:
            kwargs['zeros'] = True
            
        if self.masked:
            gnm = MaskedGNM(self._title, self.mask)
        else:
            gnm = GNM(self._title)
        gnm.setKirchhoff(self.getKirchhoff())
        gnm.calcModes(n_modes=n_modes, **kwargs)
        return gnm
    
    def normalize(self, method=VCnorm, **kwargs):
        """Applies chosen normalization on the current Hi-C map."""

        M = self._map
        N = method(M, **kwargs)
        self.map = N
        return N
    
    def setDomains(self, labels, **kwargs):
        """Uses spectral clustering to identify structural domains on the chromosome.
        
        :arg labels: domain labels
        :type labels: :class:`~numpy.ndarray`, list

        :arg method: Label assignment algorithm used after Laplacian embedding.
        :type method: func
        """
        wastrimmed = self.masked

        self.masked = True
        if len(labels) == self.numAtoms():
            full_length = self.numAtoms()
            if full_length != len(labels):
                _labels = np.empty(full_length)
                _labels.fill(np.nan)
                _labels[self.mask] = labels

                currlbl = labels[0]

                for i in range(len(_labels)):
                    l = _labels[i]
                    if np.isnan(l):
                        _labels[i] = currlbl
                    elif currlbl != l:
                        currlbl = l
                labels = _labels
        else:
            self.masked = False
            if len(labels) != self.numAtoms():
                raise ValueError('The length of the labels should match either the length '
                                 'of masked or complete Hi-C map. Turn off "masked" if '
                                 'you intended to set the labels to the full map.')
        
        self.masked = wastrimmed
        self._labels = labels
        return self.getDomains()
    
    def getDomains(self):
        """Returns an 1D :class:`numpy.ndarray` whose length is the number of loci. Each 
        element is an index denotes to which domain the locus belongs."""

        lbl = self._labels
        mask = self.mask
        if self.masked:
            lbl = lbl[mask]
        return lbl

    def getDomainList(self):
        """Returns a list of domain separations. The list has two columns: the first is for 
        the domain starts and the second is for the domain ends."""

        indicators = np.diff(self.getDomains())
        indicators = np.append(1., indicators)
        indicators[-1] = 1
        sites = np.where(indicators != 0)[0]
        starts = sites[:-1]
        ends = sites[1:]
        domains = np.array([starts, ends]).T

        return domains

    def view(self, spec='p', **kwargs):
        """Visualization of the Hi-C map and domains (if present). The function makes use 
        of :func:`.showMatrix`.
        
        :arg spec: a string specifies how to preprocess the matrix. Blank for no preprocessing,
                'p' for showing only data from *p*-th to *100-p*-th percentile. '_' is to suppress 
                creating a new figure and paint to the current one instead. The letter specifications 
                can be applied sequentially, e.g. 'p_'.
        :type spec: str

        :arg p: specifies the percentile threshold.
        :type p: double
        """

        dm_kwargs = {}
        keys = list(kwargs.keys())
        for k in keys:
            if k.startswith('dm_'):
                dm_kwargs[k[3:]] = kwargs.pop(k)
            elif k.startswith('domain_'):
                dm_kwargs[k[7:]] = kwargs.pop(k)

        M = self.map
        if 'p' in spec:
            p = kwargs.pop('p', 5)
            lp = kwargs.pop('lp', p)
            hp = kwargs.pop('hp', 100-p)
            vmin = np.percentile(M, lp)
            vmax = np.percentile(M, hp)
        else:
            vmin = vmax = None

        if not 'vmin' in kwargs:
            kwargs['vmin'] = vmin
        if not 'vmax' in kwargs:
            kwargs['vmax'] = vmax

        im = showMatrix(M, **kwargs)

        domains = self.getDomainList()
        if len(domains) > 1:
            showDomains(domains, **dm_kwargs)

        return im

    def copy(self):
        new = type(self)()
        new.__dict__.update(self.__dict__)
        return new
    
    __copy__ = copy


def parseHiC(filename, **kwargs):
    """Returns an :class:`.HiC` from a Hi-C data file.

    This function extends :func:`.parseHiCStream`.

    :arg filename: the filename to the Hi-C data file.
    :type filename: str
    """

    import os, struct
    title = kwargs.get('title')
    if title is None:
        title = os.path.basename(filename)
    else:
        title = kwargs.pop('title')

    if isURL(filename):
        M, res = parseHiCBinary(filename, title=title, **kwargs)
    else:
        with open(filename,'rb') as req:
            magic_number = struct.unpack('<3s',req.read(3))[0]
        if magic_number == b"HIC":
            M, res = parseHiCBinary(filename, title=title, **kwargs)
        else:
            with open(filename, 'r') as filestream:
                M, res = parseHiCStream(filestream, title=title, **kwargs)
    
    hic = HiC(title=title, map=M, bin=res)

    return hic

def _sparse2dense(I, J, values, bin=None):
    I = np.asarray(I, dtype=int)
    J = np.asarray(J, dtype=int)
    values = np.asarray(values, dtype=float)
    # determine the bin size by the most frequent interval
    if bin is None:
        loci = np.unique(np.sort(I))
        bins = np.diff(loci)
        bin = mode(bins)[0][0]
    # convert coordinate from basepair to locus index
    bin = int(bin)
    I = I // bin
    J = J // bin
    # make sure that the matrix is square
    # if np.max(I) != np.max(J):
    #     b = np.max(np.append(I, J))
    #     I = np.append(I, b)
    #     J = np.append(J, b)
    #     values = np.append(values, 0.)
    # Convert to sparse matrix format, then full matrix format
    # and finally array type. Matrix format is avoided because
    # diag() won't work as intended for Matrix instances.
    M = np.array(coo_matrix((values, (I, J))).todense())
    return M, bin

def parseHiCStream(stream, **kwargs):
    """Returns an :class:`.HiC` from a stream of Hi-C data lines.

    :arg stream: Anything that implements the method ``read``, ``seek``
        (e.g. :class:`file`, buffer, stdin)
    """

    issparse = kwargs.get('sparse', None)

    import csv
    dialect = csv.Sniffer().sniff(stream.read(1024))
    stream.seek(0)
    reader = csv.reader(stream, dialect)
    D = list()
    for row in reader:
        d = list()
        for element in row:
            d.append(np.double(element))
        D.append(d)
    D = np.array(D)

    res = kwargs.get('bin', None)
    if res is not None:
        res = int(res)
    size = D.shape
    if len(D.shape) <= 1:
        raise ValueError("cannot parse the file: input file only contains one column.")
    
    if issparse is None:
        issparse = size[1] == 3

    if not issparse:
        M = D
    else:
        try:
            I, J, values = D.T[:3]
        except ValueError:
            raise ValueError('the sparse matrix format should have three columns')
        
        M, res = _sparse2dense(I, J, values, bin=res)
    return M, res

def parseHiCBinary(filename, **kwargs):

    chrloc = kwargs.get('chrom', None)
    if chrloc is None:
        raise ValueError('chrom needs to be specified when parsing .hic format')
    chrloc1 = kwargs.get('chrom1', chrloc)
    chrloc2 = kwargs.get('chrom2', chrloc)
    norm = kwargs.get('norm', 'NONE')
    unit = kwargs.get('unit', 'BP')
    res = kwargs.get('binsize', None)
    res = kwargs.get('bin', res)
    if res is None:
        raise ValueError('bin needs to be specified when parsing .hic format')
    res = int(res)

    from .straw import straw
    result = straw(norm, filename, chrloc1, chrloc2, unit, res)

    M, res = _sparse2dense(*result, bin=res)
    return M, res

def writeMap(filename, map, bin=None, format='%f'):
    """Writes *map* to the file designated by *filename*.

    :arg filename: the file to be written.
    :type filename: str

    :arg map: a Hi-C contact map.
    :type map: :class:`numpy.ndarray`

    :arg bin: bin size of the *map*. If bin is `None`, *map* will be 
              written in full matrix format.
    :type bin: int

    :arg format: output format for map elements.
    :type format: str
    """

    assert isinstance(map, np.ndarray), 'map must be a numpy.ndarray.'

    if bin is None:
        return writeArray(filename, map, format=format)
    else:
        L = int(map.size - np.diag(map).size)//2 + np.diag(map).size
        spmat = np.zeros((L, 3))
        m,n = map.shape
        l = 0
        for i in range(m):
            for j in range(i,n):
                spmat[l, 0] = i * bin
                spmat[l, 1] = j * bin
                spmat[l, 2] = map[i, j]
                l += 1
        fmt = ['%d', '%d', format]
        return writeArray(filename, spmat, format=fmt)

def saveHiC(hic, filename=None, map=True, **kwargs):
    """Saves *HiC* model data as :file:`filename.hic.npz`. If *map* is **True**, 
    Hi-C contact map will not be saved and it can be loaded from raw data file 
    later. If *filename* is **None**, name of the Hi-C instance will be used as 
    the filename, after ``" "`` (white spaces) in the name are replaced with 
    ``"_"`` (underscores). Upon successful completion of saving, filename is 
    returned. This function makes use of :func:`numpy.savez` function."""

    assert isinstance(hic, HiC), 'hic must be a HiC instance.'
    
    if filename is None:
        filename = hic.getTitle().replace(' ', '_')
    
    if filename.lower().endswith('.hic'):
        filename += '.npz'
    elif not filename.lower().endswith('.hic.npz'):
        filename += '.hic.npz'

    attr_dict = hic.__dict__.copy()
    if not map:
        attr_dict.pop('_map')

    ostream = openFile(filename, 'wb', **kwargs)
    np.savez_compressed(ostream, **attr_dict)
    ostream.close()

    return filename

def loadHiC(filename):
    """Returns HiC instance after loading it from file (*filename*).
    This function makes use of :func:`numpy.load` function. See also 
    :func:`saveHiC`."""
    
    attr_dict = np.load(filename)
    hic = HiC()

    keys = attr_dict.keys()

    for k in keys:
        val = attr_dict[k]
        if len(val.shape) == 0:
            val = np.asscalar(val)
        setattr(hic, k, val)
    return hic

def saveHiC_h5(hic, filename=None, **kwargs):
    """Saves *HiC* model data as :file:`filename.hic.npz`. If *filename* is 
    **None**, name of the Hi-C instance will be used as 
    the filename, after ``" "`` (white spaces) in the name are replaced with 
    ``"_"`` (underscores). Upon successful completion of saving, filename is 
    returned. This function makes use of :func:`numpy.savez` function."""

    try:
        import h5py
    except:
        raise ImportError('h5py needs to be installed for using this function')

    assert isinstance(hic, HiC), 'hic must be a HiC instance.'
    
    if filename is None:
        filename = hic.getTitle().replace(' ', '_')
    
    if filename.lower().endswith('.hic'):
        filename += '.hic'
    elif not filename.lower().endswith('.hic.h5'):
        filename += '.hic.h5'

    attr_dict = hic.__dict__.copy()

    with h5py.File(filename, 'w') as f:
        for key in attr_dict:
            value = attr_dict[key]
            compression = None if np.isscalar(value) else 'gzip'
            f.create_dataset(key, data=value, compression=compression)

    return filename

def loadHiC_h5(filename):
    """Returns HiC instance after loading it from file (*filename*).
    This function makes use of :func:`numpy.load` function. See also 
    :func:`saveHiC`."""
    
    try:
        import h5py
    except:
        raise ImportError('h5py needs to be installed for using this function')

    hic = HiC()
    with h5py.File(filename, 'r') as f:
        for key in f.keys():
            try:
                value = f[key][:]
            except:
                value = f[key][()]
            setattr(hic, key, value)

    return hic



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from sphinx.application import ExtensionError

def add_ga_javascript(app, pagename, templatename, context, doctree):
    if not app.config.googleanalytics_enabled:
        return

    metatags = context.get('metatags', '')
    metatags += """<script type="text/javascript">

      var _gaq = _gaq || [];
      _gaq.push(['_setAccount', '%s']);
      _gaq.push(['_trackPageview']);

      (function() {
        var ga = document.createElement('script'); ga.type = 'text/javascript'; ga.async = true;
        ga.src = ('https:' == document.location.protocol ? 'https://ssl' : 'http://www') + '.google-analytics.com/ga.js';
        var s = document.getElementsByTagName('script')[0]; s.parentNode.insertBefore(ga, s);
      })();
    </script>""" % app.config.googleanalytics_id
    context['metatags'] = metatags

def check_config(app):
    if not app.config.googleanalytics_id:
        raise ExtensionError("'googleanalytics_id' config value must be set for ga statistics to function properly.")

def setup(app):
    app.add_config_value('googleanalytics_id', '', 'html')
    app.add_config_value('googleanalytics_enabled', True, 'html')
    app.connect('html-page-context', add_ga_javascript)
    app.connect('builder-inited', check_config)
#    return app



"""This module contains unit tests for :mod:`~prody.ensemble`."""

from prody.tests import TestCase

from numpy import ones
from numpy.testing import assert_equal

from . import ATOMS, CONF, COORDS, CONFW, ENSEMBLE
from . import PDBCONF, PDBENSEMBLE, WEIGHTS, WEIGHTS_BOOL


class TestConformation(TestCase):


    def testCoordinates(self):

        assert_equal(CONF.getCoords(), COORDS,
                     'failed to get coordinates for conformation')

    def testWeights(self):

        weights = CONFW.getWeights()
        self.assertEqual(weights.ndim, 2,
                        'wrong ndim for weights of Conformation')
        self.assertTupleEqual(weights.shape,
                              (CONFW.numAtoms(), 1),
                              'wrong shape for weights of Conformation')
        assert_equal(weights, ones((ATOMS.numAtoms(), 1), float),
                     'failed to set weights for Conformation')

    def testCoordinatesForAll(self):

        for i, conf in enumerate(ENSEMBLE):
            assert_equal(conf.getCoords(), ATOMS.getCoordsets(i),
                         'failed to get coordinates for Conformation')

    def testGetIndex(self):
        """Test get index function."""

        for i, conf in enumerate(ENSEMBLE):
            self.assertEqual(conf.getIndex(), i,
                             'failed to get correct index')

    def testGetNumAtoms(self):
        """Test get index function."""

        for i, conf in enumerate(ENSEMBLE):
            self.assertEqual(conf.numAtoms(), ATOMS.numAtoms(),
                             'failed to get correct number of atoms')


class TestPDBConformation(TestCase):


    def testCoordinates(self):

        assert_equal(PDBCONF.getCoords(), COORDS,
                     'failed to get correct coordinates')

    def testWeights(self):

        weights = PDBCONF.getWeights()
        self.assertEqual(weights.ndim, 2,
                        'wrong ndim for weights')
        self.assertTupleEqual(weights.shape, (PDBCONF.numAtoms(), 1),
                              'wrong shape for weights')
        assert_equal(PDBCONF.getWeights(), WEIGHTS[0],
                     'failed to set weights')

    def testWeightsForAll(self):

        for i, conf in enumerate(PDBENSEMBLE):
            assert_equal(conf.getWeights(), WEIGHTS[i],
                         'failed to set weights')


    def testCoordinatesForAll(self):

        for i, conf in enumerate(PDBENSEMBLE):
            assert_equal(conf.getCoords()[WEIGHTS_BOOL[i]],
                         ATOMS.getCoordsets(i)[WEIGHTS_BOOL[i]],
                         'failed to get coordinates')

    def testGetIndex(self):

        for i, conf in enumerate(PDBENSEMBLE):
            self.assertEqual(conf.getIndex(), i,
                             'failed to get correct index')

    def testGetNumAtoms(self):

        for i, conf in enumerate(PDBENSEMBLE):
            self.assertEqual(conf.numAtoms(), ATOMS.numAtoms(),
                             'failed to get correct number of atoms')



"""This module contains unit tests for :mod:`~prody.select`."""

import os
import os.path
import inspect
import numpy as np
from numpy.testing import *

from prody import *
from prody import LOGGER
from prody.tests import unittest
from prody.tests.datafiles import *
from prody.atomic.atommap import DUMMY

from prody.utilities import importDec
dec = importDec()

prody.atomic.select.DEBUG = False
LOGGER.verbosity = 'none'

TESTS_PATH = os.path.abspath(os.path.split(inspect.getfile(
                                                   inspect.currentframe()))[0])

# If a selection string is paired with None, SelectionError is expected
# If two selection strings are paired, they must select exactly same of atoms
# Else, number must be the number atoms that the string is expected to select

pdb3mht = prody.parsePDB(pathDatafile('pdb3mht.pdb'), secondary=True)

SELECTION_TESTS = {'pdb3mht':
    {'n_atoms': len(pdb3mht),
     'ag': pdb3mht,
     'all': pdb3mht.all,
     'atommap': AtomMap(pdb3mht, [DUMMY] + list(range(1500)) + [DUMMY] +
                        list(range(1500, len(pdb3mht))) + [DUMMY]),

     'test_flags':  [('none', 0),
        ('all', 3211),
        ('acidic', 334),
        ('acyclic', 2040),
        ('aliphatic', 821),
        ('aromatic', 475),
        ('at', 0),
        ('basic', 450),
        ('buried', 944),
        ('cg', 0),
        ('charged', 784),
        ('cyclic', 566),
        ('heme', 0),
        ('hydrophobic', 999),
        ('ion', 0),
        ('large', 1629),
        ('lipid', 0),
        ('medium', 689),
        ('neutral', 1822),
        ('nucleic', 509),
        ('nucleotide', 509),
        ('nucleobase', 0),
        ('nucleoside', 0),
        ('polar', 1607),
        ('protein', 2606, 'aminoacid'),
        ('stdaa', 2606),
        ('nonstdaa', 0),
        ('purine', 0),
        ('pyrimidine', 0),
        ('small', 288),
        ('sugar', 0),
        ('surface', 1662),
        ('water', 70),
        ('hetero', 96),
        ('hetatm', 96),
        ('calpha', 327, 'ca'),
        ('backbone', 1308, 'bb'),
        ('backbonefull', 1309, 'bbfull'),
        ('sidechain', 1297, 'sc'),
        ('carbon', 1920),
        ('hydrogen', 0),
        ('noh', 3211),
        ('nitrogen', 542),
        ('oxygen', 711),
        ('sulfur', 14),
        ('extended', 503),
        ('helix', 763),
        ('helix310', 118),
        ('turn', 0),
        ('bridge', 0),
        ('bend', 0),
        ('coil', 1222),],

     'test_without_and': [
        ('coil protein', 1222),
        ('sidechain sc protein', 1297),
        ('bbfull bb', 1308),
        ('(charged basic)', 450),
        ('(protein nucleic)', 0),
        ('noh hetero water', 70, 'water hetero noh'),
        ('ca occupancy > 0', 327, 'occupancy > 0 ca'),
        ('ca occupancy - 0 > 0', 327, 'occupancy - 0 > 0 ca'),
        ('ca occupancy - 0 > 0 + 0', 327,
        'occupancy - 0 > 0 + 0 ca'),
        ('occupancy > ca 0', None),
        ('noh hetero (water)', 70),
        ('noh hetero not (water)', 26),
        ('(water) hetero', 70),
        ('ca abs(beta) = beta + abs(0)', 327,
        'abs(beta) = beta + abs(0) ca'),],

     'test_unintended': [
        ('abs beta = beta', 3211, 'abs (beta) = beta')],

     'test_string': [
        ('name P', 24),
        ('name P CA', 352),
        ('name `A 1`', 0),
        ('chain C', 248),
        ('chain x', 0),
        ('chain x y', 0),
        ('chain x y z', 0),
        ('chain x y z C', 248),
        ('chain C D', 521),
        ('chain CD', 0),
        ('resname DG', 132),
        ('resname DG ALA', 212),
        ('altloc A', 0),
        ('altloc _', 3211),
        ('secondary H', 763, 'helix'),
        ('secondary H E', 1266),
        ('secondary _', 605),
        ('segment _', 3211),],

     'test_integer': [
        ('index 10 20 10000', 2),
        ('serial 0', 0),
        ('serial 1 2', 2),
        ('resnum 0', 0),
        ('resnum 100 105', 13),
        ('resid 0', 0),
        ('resid 100 105', 13),
        ('resid 100 A 105', 13),
        ('fragindex 0', None),
        ('fragment 0', None),],

     'test_range':  [
        ('index 0:10', 10),
        ('index 0to10', 11, 'index 0 to 10'),
        ('serial 0:10:2', 4),
        ('serial 0:10:10', 0),
        ('resnum 10to15', 49),
        ('resnum 10:16:1', 49),
        ('resnum `-3:16:1`', 125),
        ('resid 10to15', 49),
        ('resid 10:16:1', 49),
        ('x `-10:20`', 673, 'x `-10 to 20`'),
        ('x 0:20:1', 0),
        ('beta 13.02:13.01', None)],

     'test_float':  [
        ('beta 5.0 41.15 11.85', 2),
        ('occupancy 1.0', 3211),
        ('x 6.665', 1),
        ('y 69.99 13.314', 2),
        ('z 115.246 45.784', 2),
        ('charge 0', 0),
        ('mass 1', 0),
        ('radius 0', None),
        ('beta "1."', 0),
        ('beta = "1."', None),],

     'test_comparisons': [
        ('x = -51.659', 1),
        ('x != -51.659', 3210),
        ('z >= 82.813', 1670),
        ('z < 82.813', 1541),
        ('beta > 10', 2874),
        ('beta < 10', 336),
        ('occupancy > 0.999999', 3211),
        ('-10 <= x < 0', 557, '-10 <= x and x < 0'),
        ('11 > 10', None),
        ('radius > 10', None),
        ('chain = A', None),
        ('x x < 1', None),
        ('name < 1', None),],

     'test_operation': [
        ('x ** 2 < 10', 238),
        ('x ** 2 ** 2 ** 2 < 10', 87),
        ('x ** (+2 ** (+2 ** +2)) < 10', 87),
        ('occupancy % 2 == 1', 3211),
        ('x**2 + y**2 + z**2 < 10000', 1975),],

     'test_function': [
        ('sqrt(x**2 + y**2 + z**2) < 100', 1975,
        'x**2 + y**2 + z**2 < 10000'),
        ('sqrt(x**2 + y**2 + z**2) == '
        '(x**2 + y**2 + z**2) ** 0.5', 3211),
        ('beta % 3 < 1', 1070),
        ('beta % 4 % 3 < 1', 1530),
        ('ceil(beta) == 10', 60),
        ('floor(beta) == 10', 58),
        ('abs(x) == sqrt(sq(x))', 3211),
        ('sq(x-5)+sq(y+4)+sq(z) > sq(100)', 1444),
        ('1 > sq(occ)', None),
        ('sq(x x) > 1', None),],

     'test_composite': [
        ('chain x y z C and x 10', 0),
        ('resnum `1` `2`', 16, 'resnum 1 2'),
        ('same residue as within 4 of resname SAH', 177),
        ('name CA and same residue as within 4 of resname SAH', 20),
        ('water and within 5 of not protein', 70),
        ('backbone and sqrt((x - 25)**2 + (y - 74)**2 + '
         '(z - 13)**2) <= 500', 1308),
        ('(not resname SAH) and (protein and name CA) or '
         '(nucleic and name P)', 351,
         '(protein and name CA) or (nucleic and name P)'),
        ('protein and (backbone or name H)', 1308),
        ('same residue as within 4 of and resname SAH', None),
        ('protein and name CA CB and same residue as '
         '((x+21.2)**2 + (y-35.9)**2 + (z-80.0)**2)**0.5 < 10',
         78, 'protein and name CA CB and same residue as '
         'within 10 of center', {'center': np.array([21.2, 35.9, 80.0])})],

     'test_within': [
        ('within 10 of index 0', 72),
        ('exwithin 100 of index 0', 3210),
        ('exwithin 4 of resname SAH', 61),
        ('(within 4 of water) and not water', 534, 'exwithin 4 of water'),
        ('within 5 of within 5 of within 5 of index 0', 135),
        ('exwithin 5 of exwithin 5 of exwithin 5 of index 0', 99),
        ('within 1 of pdb', 3211, None, {'pdb': pdb3mht}),
        ('exwithin 1 of pdb', 0, None, {'pdb': pdb3mht}),
        ('exwithin 1 of ag', None, None, {'ag': AtomGroup()}),
        ('within 100 of index 10000', 0),],

     'test_sameas': [
         ('same residue as index 0', 22),
         ('same chain as index 0', 248),
         ('same segment as index 0', 3211),
         ('same residue as resname DG ALA', 212),
         ('same chain as chain C', 248),
         ('same residue as chain X', 0),
         ('same none as chain C', None),
         ('same residue as same residue as same residue as index 0', 22,
          'resindex 0')],

     'test_regexp': [
        ('resname "S.."', 122),
        ('name "C.*"', 1920),
        ('name ".*\'"', 208),
        ('name "C(A|B)"', 628),
        ('name "C((A|B)"', None),],

     'test_specialchar': [
        ('altloc ` `', 3211),
        ('name A` `CA`', 328),
        ('name `A``', 0),
        ('z `+100.291`', 1),],

     'test_logical': [
        ('name or name', None),
        ('name and name', None),
        ('name CA and name CA', 328),
        ('name CA or name CA', 328),
        ('index 0 or index 1 ', 2),
        ('not not not not index 1', 1),
        ('index 0 or index 1 or index 2', 3, 'index 0 1 2'),
        ('index 0 or index 1 or index 2 or index 4', 4, 'index 0 1 2 4'),
        ('index 0 and index 1 ', 0),
        ('index < 50 and index < 5', 5, 'index < 5'),
        ('index < 50 and index < 25 and index < 5', 5),
        ('index < 5 and index < 25 and index < 50', 5),
        ('index 0 to 5 and index 0 to 25 and index 0 to 50', 6),
        ('index < 5 and index < 25 and index < 50 or index < 50 or index < 5',
         50),],

     'test_kwargs': [
        ('within 100 of origin', 1975, None, {'origin': np.zeros(3)}),
        ('within 100 of origin', 1975, None, {'origin': np.zeros((1, 3))}),
        ('within 100 of origin', 1975, None, {'origin': np.zeros((10, 3))}),
        ('within 100 of origin', 1975, None, {'origin': np.zeros((50, 3))}),
        ('within 100 of none', None, None, {'none': np.zeros((50, 3))}),],

     'test_equivalent':[
        ('chain C', 248, 'not not chain C'),
        ('chain C', 248, 'not not not not chain C'),
        ('nucleic', 509, 'nucleoside or nucleotide or nucleobase'),],

     'test_invalid':[
        ('chain C and and chain C', None),
        ('chain C or or chain D', None),
        ('chain C or not or chain D', None),
        ('chain C + 3', None),
        ('sqr(x-5)+sqr(y+4)+sqr(z) > sqr(100)', None),
        ('x > sq(calpha)', None),
        ('x > sq(name CA and resname ALA)', None),
        ('resname ALA and +1', None)],

     'test_userdata':[
        ('temp < 10', 336, 'beta < 10'),
        ('temp < 10 and chain D', 37, 'temp < 10 and chain D'),
        ('oc10 - 9 == 1', 3211, 'occupancy 1'),
        ('temp < 10', 336, 'temp + oc10 < 20'),
        ('occ', 3211, 'occupancy != 0'),
        ('occ and occupancy == 1', 3211, 'occupancy != 0'),
        ('occ and occupancy == 1 and oc10 - 9 == 1', 3211),
        ('occ and occupancy == 1 and temp < 10', 336),
        ('occ > 0', None),],

     'test_synonyms':    [
        ('chain C', 248, 'chid C'),
        ('chain C D', 521, 'chid C D'),],

     'test_sequence': [
        ('sequence al', 0),
        ('sequence A', 80, 'resname ALA'),
        ('sequence MIEIK', 42, 'resindex 25 to 29'),
        ('sequence VLNAL', 36, 'resindex 175 to 179'),
        ('sequence FKPY', 40, 'resindex 348 to 351'),
        ('sequence "SS."', 20, 'resindex 344 to 346'),
        ('sequence "S[A-Z]{2}G"', 79,
         'resindex 109 to 112 267 to 270 276 to 279'),
        ('sequence "S.S.S"', 0),
        ('sequence "."', 2606),],

     'test_docexamples': [
        ('serial 1 2 3', 3),
        ('serial 1 to 10', 10),
        ('serial 1:10:2', 5),
        ('serial < 10', 9),
        ('beta 555.55', 0),
        ('beta 1 to 500', 3211),
        ('beta 1:500', 3211),
        ('beta < 500', 3211),
        ('resnum 120A 120B', 0),
        ('icode A', 0),
        ('icode _', 3211),
        ('charge 1', 3211),
        ('abs(charge) == 1', 3211),
        ('charge < 0', 0),
        ('0 < mass < 500', 3211),
        ('abs(mass) <= mass <= 10', 337),],
    }
}

subsets = []
for ch in pdb3mht.iterChains():
    subsets.append((ch.getSelstr(), ch.numAtoms()))

for i, res in enumerate(pdb3mht.iterResidues()):
    if i % 80 == 0:
        subsets.append((res.getSelstr(), res.numAtoms()))

SELECTION_TESTS['pdb3mht']['subsets'] = subsets

ligand = fetchPDBLigand(pathDatafile('sti'))['ideal']

SELECTION_TESTS['imatinib'] = {
    'n_atoms': len(ligand),
    'ag': ligand,
    'all': ligand.all,
    'atommap': AtomMap(ligand, [DUMMY] + list(range(10)) + [DUMMY] +
                       list(range(10, len(ligand))) + [DUMMY]),

    'test_bondedto': [
        ('bonded to index 0', ligand[0].numBonds() + 1),
        ('exbonded to index 0', ligand[0].numBonds()),
        ('bonded to index 67', ligand[67].numBonds() + 1),
        ('exbonded to index 67', ligand[67].numBonds()),
        ('bonded 2 to index 0', 8, 'bonded to bonded to index 0'),
        ('bonded 0 to index 0', 0),
        ('bonded 3 to index 0', 10,
         'bonded to bonded to bonded to index 0'),
        ('bonded 4 to index 0', 13,
         'bonded to bonded to bonded to bonded to index 0'),
        ('bonded 4 to index 0', 13,
         'bonded 2 to bonded 2 to index 0'),
        ('bonded 4 to index 0', 13,
         'bonded to bonded 3 to index 0'),
        ('exbonded 1 to index 0', 3, 'exbonded to index 0'),
        ('exbonded 2 to index 0', 5,
         'exbonded to exbonded to index 0'),
        ('exbonded 3 to index 0', 5,
         'exbonded to exbonded to exbonded to index 0'),
        ('bonded 20 to index 0', 64),
        ('bonded 20 to index 10', 68),
        ('bonded to index 1000', 0),
        ('fragment 0', len(ligand)),
        ('fragment 1', 0),
        ('fragment 0 1 2', len(ligand)),
        ('fragindex 0 1 2', len(ligand)),
        ('fragindex 0:2', len(ligand)),],
}

pdb_gromacs = prody.parsePDB(pathDatafile('pdb6fpj_Bb_fixed_solv_ions.pdb'), secondary=True)
SELECTION_TESTS['gromacs'] = {
    'n_atoms': len(pdb_gromacs),
    'ag': pdb_gromacs,
    'all': pdb_gromacs.all,
    'test_flags':  [('protein', 12194, 'aminoacid')],
}

data_nucleoside = DATA_FILES['pdb7pbl']
pdb_nucleoside = prody.parsePDB(pathDatafile(data_nucleoside['file']), secondary=True)
SELECTION_TESTS['nucleoside'] = {
    'n_atoms': data_nucleoside['atoms'],
    'ag': pdb_nucleoside,
    'all': pdb_nucleoside.all,
    'test_flags':  [('nucleoside', data_nucleoside['nucleoside'], 'nucleoside')],
}

pdb3mht = SELECTION_TESTS['pdb3mht']['ag']
pdb3mht.setCharges(pdb3mht.getOccupancies())
pdb3mht.setMasses(pdb3mht.getBetas())
pdb3mht.setData('temp', pdb3mht.getBetas())
pdb3mht.setData('oc10', pdb3mht.getOccupancies() * 10)
pdb3mht.setFlags('occ', pdb3mht.getOccupancies().astype(bool))

SELECT = prody.Select()

EMPTYDICT = {}


class TestSelect(unittest.TestCase):

    """Test :class:`.Select`."""

    pass

count = 0
for case, items in SELECTION_TESTS.items():

    for key, tests in items.items():

        if not key.startswith('test_'):
            continue
        type_ = key[5:]

        for test in tests:
            def func(self, pdb=case, test=test, type_=type_, **kwargs):

                atoms = SELECTION_TESTS[pdb]['ag']

                selstr = test[0]
                natoms = test[1]
                selstr2 = None
                kwargs = EMPTYDICT
                if len(test) == 3:
                    selstr2 = test[2]
                if len(test) == 4:
                    kwargs = test[3]

                if natoms is None:
                    self.assertRaises(prody.select.SelectionError,
                        SELECT.getIndices, atoms, selstr, **kwargs)
                elif selstr2 is None:
                    sel = SELECT.getIndices(atoms, selstr, **kwargs)
                    self.assertEqual(len(sel), natoms,
                        'selection {0} for {1} failed, expected '
                        '{2}, selected {3}'.format(repr(selstr),
                        str(atoms), natoms, len(sel)))
                else:
                    sel = SELECT.getIndices(atoms, selstr, **kwargs)
                    sel2 = SELECT.getIndices(atoms, selstr2, **kwargs)
                    self.assertTrue(len(sel) == len(sel2) == natoms and
                            np.all(sel == sel2),
                        'selection strings {0} and {1} for '
                        '{2} failed to select same number of atoms, '
                        'expected ({3})'.format(repr(selstr),
                        repr(selstr2), str(atoms), natoms))

            count += 1
            func.__name__ = 'test{0}Selection{1}'.format(
                                                type_.title(), count)
            func.__doc__ = ('Test {0} selections {1} for '
                                    '{2}').format(type_,
                                    repr(test[0]), case)
            setattr(TestSelect, func.__name__, func)

            @dec.slow
            def func(self, pdb=case, test=test, type_=type_, **kwargs):

                atoms = SELECTION_TESTS[pdb]['all']

                selstr = test[0]
                natoms = test[1]
                selstr2 = None
                kwargs = EMPTYDICT
                if len(test) == 3:
                    selstr2 = test[2]
                if len(test) == 4:
                    kwargs = test[3]

                if natoms is None:
                    self.assertRaises(prody.select.SelectionError,
                        SELECT.getIndices, atoms, selstr, **kwargs)
                elif selstr2 is None:
                    sel = SELECT.getIndices(atoms, selstr, **kwargs)
                    self.assertEqual(len(sel), natoms,
                        'selection {0} for {1} failed, expected '
                        '{2}, selected {3}'.format(repr(selstr),
                        str(atoms), natoms, len(sel)))
                else:
                    sel = SELECT.getIndices(atoms, selstr, **kwargs)
                    sel2 = SELECT.getIndices(atoms, selstr2, **kwargs)
                    self.assertTrue(len(sel) == len(sel2) == natoms and
                            np.all(sel == sel2),
                        'selection strings {0} and {1} for '
                        '{2} failed to select same number of atoms, '
                        'expected ({3})'.format(repr(selstr),
                        repr(selstr2), str(atoms), natoms))

            count += 1
            func.__name__ = 'test{0}Selection{1}'.format(type_.title(),
                                                             count)
            func.__doc__ = 'Test {0} selections "{1}"'.format(type_,
                                                                  test[0])
            setattr(TestSelect, func.__name__, func)
del func

MACROS = [('cacb', 'name CA CB'),
          ('donors', '(protein) and (name N NE NH2 ND2 NE2 ND1 OG OH NH1 '
                                         'SG OG1 NE2 NZ NE1 ND1 NE2)')]


class TestMacros(unittest.TestCase):

    """Test selection macros."""


    def testMacroFunctions(self):

        for name, macro in MACROS:
            prody.defSelectionMacro(name, macro)
            self.assertEqual(prody.getSelectionMacro(name), macro,
                             'failed to get correct macro definition')
            prody.delSelectionMacro(name)

count = 0
for name, macro in MACROS:

    def func(self, name=name, macro=macro):

        prody.defSelectionMacro(name, macro)
        for key, case in SELECTION_TESTS.items():
            atoms = case['ag']
            assert_equal(
                SELECT.getBoolArray(atoms, macro),
                SELECT.getBoolArray(atoms, name),
                'failed to select correct selection using macro')
        prody.delSelectionMacro(name)
    count += 1

    func.__name__ = 'testMacro{0}'.format(count)
    func.__doc__ = 'Test macro *{0}*: {1}'.format(name, repr(macro))
    setattr(TestMacros, func.__name__, func)
del func


def testGetBoolArray():

    ca = pdb3mht.ca
    assert_equal(len(ca), len(SELECT.getBoolArray(ca, 'index 510')))




# -*- coding: utf-8 -*-
"""This module defines a function for parsing protein structure files in
`PSF format`_.

.. _PSF format:
   http://www.ks.uiuc.edu/Training/Tutorials/namd/
   namd-tutorial-unix-html/node23.html"""

import os.path

from numpy import fromstring, zeros, ones, array, add

from prody import PY2K
from prody.atomic import ATOMIC_FIELDS, AtomGroup
from prody.utilities import openFile

if PY2K:
    range = xrange

__all__ = ['parsePSF', 'writePSF']

def parsePSF(filename, title=None, ag=None):
    """Returns an :class:`.AtomGroup` instance storing data parsed from X-PLOR
    format PSF file *filename*.  Atom and bond information is parsed from the
    file.  If *title* is not given, *filename* will be set as the title of the
    :class:`.AtomGroup` instance.  An :class:`.AtomGroup` instance may be
    provided as *ag* argument.  When provided, *ag* must have the same number
    of atoms in the same order as the file.  Data from PSF file will be added
    to the *ag*.  This may overwrite present data if it overlaps with PSF file
    content.  
    
    This function now includes the angles, dihedrals, and impropers sections
    as well as donors, acceptors and crossterms!"""

    if ag is not None:
        if not isinstance(ag, AtomGroup):
            raise TypeError('ag must be an AtomGroup instance')

    psf = openFile(filename, 'rb')
    line = psf.readline()
    while line:
        line = line.strip()
        if line.endswith(b'!NATOM'):
            n_atoms = int(line.split(b'!')[0])
            break
        line = psf.readline()
    if title is None:
        title = os.path.splitext(os.path.split(filename)[1])[0]
    else:
        title = str(title)
    if ag is None:
        ag = AtomGroup(title)
    else:
        if n_atoms != ag.numAtoms():
            raise ValueError('ag and PSF file must have same number of atoms')

    serials = zeros(n_atoms, ATOMIC_FIELDS['serial'].dtype)
    segnames = zeros(n_atoms, ATOMIC_FIELDS['segment'].dtype)
    resnums = zeros(n_atoms, ATOMIC_FIELDS['resnum'].dtype)
    resnames = zeros(n_atoms, ATOMIC_FIELDS['resname'].dtype)
    atomnames = zeros(n_atoms, ATOMIC_FIELDS['name'].dtype)
    atomtypes = zeros(n_atoms, ATOMIC_FIELDS['type'].dtype)
    charges = zeros(n_atoms, ATOMIC_FIELDS['charge'].dtype)
    masses = zeros(n_atoms, ATOMIC_FIELDS['mass'].dtype)
    
    n = 0
    n_bonds = 0
    for line in psf:
        if line.strip() == b'':
            continue
        if b'!NBOND:' in line.upper():
            items = line.split()
            n_bonds = int(items[0])
            break
        if n + 1 > n_atoms:
            continue

        if len(line) <= 71:
            serials[n] = line[:8]
            segnames[n] = line[9:13].strip()
            resnums[n] = line[14:19]
            resnames[n] = line[19:23].strip()
            atomnames[n] = line[24:28].strip()
            atomtypes[n] = line[29:35].strip()
            charges[n] = line[35:44]
            masses[n] = line[50:60]
        else:
            items = line.split()
            serials[n] = items[0]
            segnames[n] = items[1]
            resnums[n] = items[2]
            resnames[n] = items[3]
            atomnames[n] = items[4]
            atomtypes[n] = items[5]
            charges[n] = items[6]
            masses[n] = items[7]
        n += 1
    
    if n < n_atoms:
        raise IOError('number of lines in PSF atoms block is less than the number of '
                      'atoms')

    n_angles = 0
    lines = []
    for i, line in enumerate(psf):
        if line.strip() == b'':
            continue
        if b'!NTHETA' in line:
            items = line.split()
            n_angles = int(items[0])
            break
        lines.append(line.decode(encoding='UTF-8'))
    
    lines = ''.join(lines)
    b_array = fromstring(lines, count=n_bonds*2, dtype=int, sep=' ')
    if len(b_array) != n_bonds*2:
        raise IOError('number of bonds expected and parsed do not match')

    n_dihedrals = 0
    lines = []
    for i, line in enumerate(psf):
        if line.strip() == b'':
            continue
        if b'!NPHI' in line:
            items = line.split()
            n_dihedrals = int(items[0])
            break
        lines.append(line.decode(encoding='UTF-8'))
    
    lines = ''.join(lines)
    a_array = fromstring(lines, count=n_angles*3, dtype=int, sep=' ')
    if len(a_array) != n_angles*3:
        raise IOError('number of angles expected and parsed do not match')

    n_impropers = 0
    lines = []
    for i, line in enumerate(psf):
        if line.strip() == b'':
            continue
        if b'!NIMPHI' in line:
            items = line.split()
            n_impropers = int(items[0])
            break
        lines.append(line.decode(encoding='UTF-8'))
    
    lines = ''.join(lines)
    d_array = fromstring(lines, count=n_dihedrals*4, dtype=int, sep=' ')
    if len(d_array) != n_dihedrals*4:
        raise IOError('number of dihedrals expected and parsed do not match')

    n_donors = 0
    lines = []
    for i, line in enumerate(psf):
        if line.strip() == b'':
            continue
        if b'!NDON' in line:
            items = line.split()
            n_donors = int(items[0])
            break
        lines.append(line.decode(encoding='UTF-8'))
    
    lines = ''.join(lines)
    i_array = fromstring(lines, count=n_impropers*4, dtype=int, sep=' ')
    if len(i_array) != n_impropers*4:
        raise IOError('number of impropers expected and parsed do not match')

    n_acceptors = 0
    lines = []
    for i, line in enumerate(psf):
        if line.strip() == b'':
            continue
        if b'!NACC' in line:
            items = line.split()
            n_acceptors = int(items[0])
            break
        lines.append(line.decode(encoding='UTF-8'))
    
    lines = ''.join(lines)
    do_array = fromstring(lines, count=n_donors*2, dtype=int, sep=' ')
    if len(do_array) != n_donors*2:
        raise IOError('number of donors expected and parsed do not match')

    n_exclusions = 0
    lines = []
    for i, line in enumerate(psf):
        if line.strip() == b'':
            continue
        if b'!NNB' in line:
            items = line.split()
            n_exclusions = int(items[0])
            break
        lines.append(line.decode(encoding='UTF-8'))
    
    lines = ''.join(lines)
    ac_array = fromstring(lines, count=n_acceptors*2, dtype=int, sep=' ')
    if len(ac_array) != n_acceptors*2:
        raise IOError('number of acceptors expected and parsed do not match')

    lines = []
    for i, line in enumerate(psf):
        if line.strip() == b'':
            continue
        if b'!' in line:
            break
        lines.append(line.decode(encoding='UTF-8'))
    
    lines = ''.join(lines)
    nbe_array = fromstring(lines, count=n_exclusions*2, dtype=int, sep=' ')
    if len(nbe_array) != n_exclusions*2:
        raise IOError('number of nonbonded exclusions expected and parsed do not match')

    n_crossterms = 0
    for i, line in enumerate(psf):
        if b'!NCRTERM' in line:
            items = line.split()
            n_crossterms = int(items[0])
            break

    lines = []
    for i, line in enumerate(psf):
        lines.append(line.decode(encoding='UTF-8'))
    
    lines = ''.join(lines)
    c_array = fromstring(lines, count=n_crossterms*4, dtype=int, sep=' ')
    if len(c_array) != n_crossterms*4:
        raise IOError('number of crossterms expected and parsed do not match')

    psf.close()
    ag.setSerials(serials)
    ag.setSegnames(segnames)
    ag.setResnums(resnums)
    ag.setResnames(resnames)
    ag.setNames(atomnames)
    ag.setTypes(atomtypes)
    ag.setCharges(charges)
    ag.setMasses(masses)

    if n_bonds > 0:
        b_array = add(b_array, -1, b_array)
        ag.setBonds(b_array.reshape((n_bonds, 2)))

    if n_angles > 0:
        a_array = add(a_array, -1, a_array)
        ag.setAngles(a_array.reshape((n_angles, 3)))

    if n_dihedrals > 0:
        d_array = add(d_array, -1, d_array)
        ag.setDihedrals(d_array.reshape((n_dihedrals, 4)))

    if n_impropers > 0:
        i_array = add(i_array, -1, i_array)
        ag.setImpropers(i_array.reshape((n_impropers, 4)))

    if n_donors > 0:
        do_array = add(do_array, -1, do_array)
        ag.setDonors(do_array.reshape((n_donors, 2)))

    if n_acceptors > 0:
        ac_array = add(ac_array, -1, ac_array)
        ag.setAcceptors(ac_array.reshape((n_acceptors, 2)))

    if n_exclusions > 0:
        nbe_array = add(nbe_array, -1, nbe_array)
        ag.setNBExclusions(nbe_array.reshape((n_exclusions, 2)))

    if n_crossterms > 0:
        c_array = add(c_array, -1, c_array)
        ag.setCrossterms(c_array.reshape((n_crossterms, 4)))

    return ag


PSFLINE = ('%8d %-4s %-4d %-4s %-4s %-4s %10.6f %13.4f %11d\n')

def writePSF(filename, atoms):
    """Write atoms in X-PLOR format PSF file with name *filename* and return
    *filename*.  This function will write available atom and bond information
    only."""

    if not filename.lower().endswith('.psf'):
        filename = filename + '.psf'

    try:
        n_atoms, segments, rnums, rnames, names, types, charges, masses = (
        atoms.numAtoms(), atoms._getSegnames(), atoms._getResnums(),
        atoms._getResnames(), atoms._getNames(), atoms._getTypes(),
        atoms._getCharges(), atoms._getMasses())

    except AttributeError:
        raise TypeError('atoms must be an Atomic instance')

    if segments is None:
        segments = atoms._getChids()
        if segments is None:
            segments = ['UNK'] * n_atoms

    if rnums is None:
        rnums = ones(n_atoms, int)

    if rnames is None:
        rnames = ['UNK'] * n_atoms

    if names is None:
        raise ValueError('atom names are not set')

    if types is None:
        types = zeros(n_atoms, ATOMIC_FIELDS['type'].dtype)

    if charges is None:
        charges = zeros(n_atoms, ATOMIC_FIELDS['charge'].dtype)

    if masses is None:
        masses = zeros(n_atoms, ATOMIC_FIELDS['mass'].dtype)

    long_fields = array([len(tp) for tp in types]).max() > 4

    out = openFile(filename, 'w')
    write = out.write
    write('PSF{0}\n'.format( ' NAMD' if long_fields else ''))
    write('\n')
    write('{0:8d} !NTITLE\n'.format(1))
    write(' REMARKS {0}\n'.format(str(atoms)))
    write('\n')
    write('{0:8d} !NATOM\n'.format(n_atoms))

    for i in range(n_atoms):
        write(PSFLINE % (i + 1, segments[i], rnums[i], rnames[i], names[i],
                        types[i], charges[i], masses[i], 0))

    bonds = list(atoms._iterBonds())
    write('\n')
    write('{0:8d} !NBOND: bonds\n'.format(len(bonds)))
    if len(bonds) > 0:
        bonds = array(bonds, int) + 1
        for i, bond in enumerate(bonds):
            write('%8s%8s' % (bond[0], bond[1]))
            if i % 4 == 3:
                write('\n')
        if i % 4 != 3:
            write('\n')

    write('\n')

    angles = list(atoms._iterAngles())
    write('\n')
    write('{0:8d} !NTHETA: angles\n'.format(len(angles)))
    if len(angles) > 0:
        angles = array(angles, int) + 1
        for i, angle in enumerate(angles):
            write('%8s%8s%8s' % (angle[0], angle[1], angle[2]))
            if i % 3 == 2:
                write('\n')
        if i % 3 != 2:
            write('\n')

    write('\n')

    dihedrals = list(atoms._iterDihedrals())
    write('\n')
    write('{0:8d} !NPHI: dihedrals\n'.format(len(dihedrals)))
    if len(dihedrals) > 0:
        dihedrals = array(dihedrals, int) + 1
        for i, dihedral in enumerate(dihedrals):
            write('%8s%8s%8s%8s' % (dihedral[0], dihedral[1], dihedral[2], dihedral[3]))
            if i % 4 == 3:
                write('\n')
        if i % 4 != 3:
            write('\n')

    write('\n')

    impropers = list(atoms._iterImpropers())
    write('\n')
    write('{0:8d} !NIMPHI: impropers\n'.format(len(impropers)))
    if len(impropers) > 0:
        impropers = array(impropers, int) + 1
        for i, improper in enumerate(impropers):
            write('%8s%8s%8s%8s' % (improper[0], improper[1], improper[2], improper[3]))
            if i % 2 == 1:
                write('\n')
        if i % 2 != 1:
            write('\n')

    write('\n')

    donors = list(atoms._iterDonors())
    write('\n')
    write('{0:8d} !NDON: donors\n'.format(len(donors)))
    if len(donors) > 0:
        donors = array(donors, int) + 1
        for i, donor in enumerate(donors):
            write('%8s%8s' % (donor[0], donor[1]))
            if i % 4 == 3:
                write('\n')
        if i % 4 != 3:
            write('\n')
    
    write('\n')

    acceptors = list(atoms._iterAcceptors())
    write('\n')
    write('{0:8d} !NACC: acceptors\n'.format(len(acceptors)))
    if len(acceptors) > 0:
        acceptors = array(acceptors, int) + 1
        for i, acceptor in enumerate(acceptors):
            write('%8s%8s' % (acceptor[0], acceptor[1]))
            if i % 4 == 3:
                write('\n')
        if i % 4 != 3:
            write('\n')

    write('\n')

    nbexclusions = list(atoms._iterNBExclusions())
    write('\n')
    write('{0:8d} !NNB\n'.format(len(nbexclusions)))
    if len(nbexclusions) > 0:
        nbexclusions = array(nbexclusions, int) + 1
        for i, nbexclusion in enumerate(nbexclusions):
            write('%8s%8s' % (nbexclusion[0], nbexclusion[1]))
            if i % 4 == 3:
                write('\n')
        if i % 4 != 3:
            write('\n')

    write('\n')

    crossterms = list(atoms._iterCrossterms())
    write('\n')
    write('{0:8d} !NCRTERM: crossterms\n'.format(len(crossterms)))
    if len(crossterms) > 0:
        crossterms = array(crossterms, int) + 1
        for i, crossterm in enumerate(crossterms):
            write('%8s%8s%8s%8s' % (crossterm[0], crossterm[1], crossterm[2], crossterm[3]))
            if i % 2 == 1:
                write('\n')
        if i % 2 != 1:
            write('\n')

    write('\n')
    out.close()
    return filename



# -*- coding: utf-8 -*-
"""This module defines classes and functions to fetch, parse, and write
structural data files associated with compounds and to access and search 
structural databases, e.g. `ProteinDataBank <http://wwpdb.org>`_.

Ligand data
===========

The following function can be used to fetch metadata on PDB ligands:

  * :func:`.fetchPDBLigand` - retrieve ligand from Ligand-Expo


BIRD data
===========

The following functions can be used to fetch metadata from the 
Biologically Interesting Reference Dictionary (BIRD) for peptide-like 
ligands:

  * :func:`.fetchBIRDviaFTP` - retrieve BIRD data from PDB
  * :func:`.parseBIRD` - parse BIRD data for a particular compound or family


CCD data
===========

The following function can be used to fetch metadata from the 
Chemical Component Dictionary (CCD) for residues within ligands:

  * :func:`.parseCCD` - parse CCD data for a particular component


Functions
===========

The following functions can be used to calculate 2D similarity using Morgan Fingerprints:

  * :func:`.calc2DSimilarity` - calculate 2D similarity for a pair of SMILES
  * :func:`.calc2DSimilarityMatrix` - calculate 2D similarity matrix for a collection of SMILES

"""

__all__ = []

from . import pdbligands
from .pdbligands import *
__all__.extend(pdbligands.__all__)

from . import bird
from .bird import *
__all__.extend(bird.__all__)

from . import ccd
from .ccd import *
__all__.extend(ccd.__all__)

from . import functions
from .functions import *
__all__.extend(functions.__all__)





"""This module defines input and output functions for BILD format for ChimeraXs.

.. _bild-format: https://rbvi.ucsf.edu/chimerax/docs/user/formats/bild.html

"""


__all__ = ['writeBILD']

import os
from os.path import join, split, splitext

import numpy as np

from prody import LOGGER
from prody.utilities import openFile, getCoords

from .nma import NMA
from .anm import ANM
from .gnm import GNM, ZERO
from .pca import PCA
from .mode import Vector, Mode
from .modeset import ModeSet

def writeBILD(filename, modes, atoms, scale=25.):
    """Returns *filename* that contains *modes* and *atoms* data in BILD format
    described at :ref:`bild-format`.  :file:`.bild` extension is appended to
    filename, if it does not have an extension. 
    
    Arrows will be scaled by the square root of the variance times *scale*.

    .. note::
       #. This function skips modes with zero eigenvalues.
       #. If a :class:`.Vector` instance is given, it will be normalized
          before it is written. It's length before normalization will be
          written as the scaling factor of the vector.
       #. It is recommended to only use one mode to avoid having lots of arrows.
    """

    if not filename.lower().endswith(".bild"):
        filename += '.bild'

    if not isinstance(modes, (NMA, ModeSet, Mode, Vector)):
        raise TypeError('modes must be NMA, ModeSet, Mode, or Vector, '
                        'not {0}'.format(type(modes)))
    if modes.numAtoms() != len(atoms):
        raise Exception('number of atoms do not match')
    out = openFile(filename, 'w')

    name = modes.getTitle()
    name = name.replace(' ', '_').replace('.', '_')
    if not name.replace('_', '').isalnum() or len(name) > 30:
        name = str(atoms)
        name = name.replace(' ', '_').replace('.', '_')
    if not name.replace('_', '').isalnum() or len(name) > 30:
        name = splitext(split(filename)[1])[0]
    out.write('.comment name {0}\n'.format(name))
    try:
        coords = getCoords(atoms)
    except:
        raise ValueError('coordinates could not be retrieved from atoms')
    if coords is None:
        raise ValueError('atom coordinates are not set')

    try:
        data = atoms.getNames()
        if data is not None:
            out.write('.comment atomnames {0}\n'.format(' '.join(data)))
    except:
        pass
    try:
        data = atoms.getResnames()
        if data is not None:
            out.write('.comment resnames {0}\n'.format(' '.join(data)))
    except:
        pass
    try:
        data = atoms.getResnums()
        if data is not None:
            out.write('.comment resids ')
            data.tofile(out, ' ')
            out.write('\n')
    except:
        pass
    try:
        data = atoms.getChids()
        if data is not None:
            out.write('.comment chainids {0}\n'.format(' '.join(data)))
    except:
        pass
    try:
        data = atoms.getSegnames()
        if data is not None:
            out.write('.comment segnames {0}\n'.format(' '.join(data)))
    except:
        pass

    try:
        data = atoms.getBetas()
        if data is not None:
            out.write('.comment bfactors ')
            data.tofile(out, ' ', '%.2f')
            out.write('\n')
    except:
        pass

    format = '{0:.3f}'.format
    out.write('.comment coordinates ')
    coords.tofile(out, ' ', '%.3f')
    out.write('\n')
    count = 0
    if isinstance(modes, Vector):
        out.write('.comment mode 1 {0:.2f} \n'.format(abs(modes)))
        normed = modes.getNormed()._getArray() * scale
        for i in range(modes.numAtoms()):
            out.write('.arrow ')
            coords[3*i:3*i+1].tofile(out, ' ', '%.3f')
            normed[3*i:3*i+1].tofile(out, ' ', '%.3f')
            out.write('\n')
        out.write('\n')
        count += 1
    else:
        if isinstance(modes, Mode):
            modes = [modes]
        else:
            LOGGER.warning('It is recommended to only provide one mode '
                'to writeBILD as otherwise there will be a mess of arrows.')
        for mode in modes:
            if mode.getEigval() < ZERO:
                continue
            out.write('.comment mode {0} {1:.2f} \n'.format(
                       mode.getIndex()+1, mode.getVariance()**0.5))
            
            arr = mode._getArray() * mode.getVariance()**0.5 * scale
            for i in range(mode.numAtoms()):
                out.write('.arrow ')
                coords[i].tofile(out, ' ', '%.3f')
                out.write(' ')
                
                arr_out = arr[3*i:3*i+3] + coords[i]
                arr_out.tofile(out, ' ', '%.3f')
                out.write('\n')
            count += 1
    if count == 0:
        LOGGER.warning('No normal mode data was written. '
                       'Given modes might have 0 eigenvalues.')
    out.close()
    return filename



# -*- coding: utf-8 -*-
"""This module defines functions for interfacing Interpro database."""

__author__ = 'James Krieger'

import json
from os.path import isfile
from prody import LOGGER, PY3K

__all__ = ['searchInterpro']

prefix = 'https://www.ebi.ac.uk/interpro/wwwapi/entry/'

def searchInterpro(query, **kwargs):
    """Returns Interpro search results in a list of dictionaries.  
    
    Matching family accessions as keys will map to various properties,
    including start and end residue positions.

    :arg query: UniProt ID or PDB identifier with or without a
        chain identifier, e.g. ``'1mkp'`` or ``'1mkpA'``.  
        UniProt ID of the specified chain, or the first
        protein chain will be used for searching the Pfam database
    :type query: str

    :arg timeout: timeout for blocking connection attempt in seconds, default
        is 60
    :type timeout: int
    """
    import requests
    
    LOGGER.timeit('_interpro')
    timeout = int(kwargs.get('timeout', 60))

    if len(query) == 4:
        url = prefix + "all/structure/pdb/" + query
        
    elif len(query) == 5:
        accession = None
        
        from prody import parsePDBHeader
        try:
            polymers = parsePDBHeader(query[:4], 'polymers')
        except Exception as err:
            raise ValueError('failed to parse header for {0} ({1})'
                             .format(query[:4], str(err)))

        chid = query[4:].upper()
        
        for poly in polymers:
            if chid and poly.chid != chid:
                continue
            for dbref in poly.dbrefs:
                if dbref.database != 'UniProt':
                    continue
                accession = dbref.accession
                LOGGER.info('UniProt accession {0} for {1} chain '
                            '{2} will be used.'
                            .format(accession, query[:4], poly.chid))
                break
            if accession is not None:
                break
            
        if accession is None:
            raise ValueError('A UniProt accession for PDB {0} could not be '
                             'parsed.'.format(repr(query)))
        else:
            url = prefix + "all/protein/uniprot/" + accession
        
    else:
        url = prefix + "all/protein/uniprot/" + query

    LOGGER.debug('Retrieving Interpro search results: ' + url)
    result = None
    sleep = 2
    while LOGGER.timing('_interpro') < timeout:
        try:
            result = requests.get(url, verify=False).content
        except Exception:
            pass
        else:
            if result not in ['PEND','RUN']:
                break
        
        sleep = 20 if int(sleep * 1.5) >= 20 else int(sleep * 1.5)
        LOGGER.sleep(int(sleep), '. Trying to reconnect...')

    if not result:
        raise IOError('Interpro search timed out or failed to parse results, '
                      ' check URL: ' + url)
    else:
        LOGGER.report('Interpro search completed in %.2fs.', '_interpro')

    if PY3K:
        result = result.decode()
        
    try:
        result = json.loads(result)
    except Exception as err:
        raise ValueError('failed to parse results as json, check URL: ' + url)

    return result["results"]



# -*- coding: utf-8 -*-
"""This module defines a class for linear discriminant analysis (LDA) calculations."""

import time

import numpy as np

from prody import LOGGER
from prody.atomic import Atomic
from prody.ensemble import Ensemble
from prody.utilities import isListLike

from .nma import NMA

__all__ = ['LDA']


class LDA(NMA):

    """A class for linear discriminant analysis (LDA) of conformational
    ensembles. See examples in :ref:`pca`."""

    def __init__(self, name='Unknown'):
        NMA.__init__(self, name)

    def calcModes(self, coordsets, labels, n_modes=None, **kwargs):
        """Calculate linear discriminant analysis modes between classes.  
        This method uses :class:`sklearn.discriminant_analysis.LinearDiscriminantAnalysis`
        on coordsets with class labels.

        *coordsets* argument may be one of the following:

        * :class:`.Atomic`
        * :class:`.Ensemble`
        * :class:`.TrajBase`
        * :class:`numpy.ndarray` with shape ``(n_csets, n_atoms, 3)``

        :arg labels: a set of labels for discriminating classes
        :type labels: :class:`~numpy.ndarray`

        :arg n_modes: number of non-zero eigenvalues/vectors to calculate,
            default is **None**,
            if **None** or ``'all'`` is given, all modes will be calculated
        :type n_modes: int

        :arg n_shuffles: number of random shuffles of labels to assess variability
        :type n_shuffles: int

        Other kwargs for the LinearDiscriminantAnalysis class can also be used. n_components defaults to n_modes
        """
        try:
            from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
        except ImportError:
            raise ImportError("Please install sklearn to use LDA")

        start = time.time()
        self._clear()
        if str(n_modes).lower() == 'all':
            n_modes = None

        if isinstance(coordsets, np.ndarray):
            if (coordsets.ndim != 3 or coordsets.shape[2] != 3 or
                    coordsets.dtype not in (np.float32, float)):
                raise ValueError('coordsets is not a valid coordinate array')
            self._coordsets = coordsets
        elif isinstance(coordsets, Atomic):
            self._coordsets = coordsets._getCoordsets()
        elif isinstance(coordsets, Ensemble):
            self._coordsets = coordsets._getCoordsets()
        else:
            raise TypeError('coordsets should be Atomic, Ensemble or numpy.ndarray, not {0}'
                            .format(type(coordsets)))

        nconfs = self._coordsets.shape[0]
        if not isListLike(labels):
            raise TypeError('labels must be either a list or a numpy.ndarray, not {0}'
                            .format(type(labels)))
        if not isinstance(labels, np.ndarray):
            labels = np.array(labels)
        if labels.ndim != 1 or len(labels) != nconfs:
            raise ValueError('labels should have same number as conformers')
        
        self._n_atoms = self._coordsets.shape[1]
        
        self._coordsets = self._coordsets.reshape(nconfs, -1)
        self._labels = labels

        quiet = kwargs.pop('quiet', False)

        self._n_components = kwargs.pop('n_components', n_modes)
        self._n_shuffles = kwargs.pop('n_shuffles', 0)
        self._lda = LinearDiscriminantAnalysis(n_components=self._n_components, **kwargs)
        self._projection = self._lda.fit(self._coordsets, self._labels)

        values = self._lda.explained_variance_ratio_
        self._eigvals = values
        self._n_modes = len(self._eigvals)
        self._vars = values

        self._array = np.array([self._lda.scalings_[:,i]/(self._lda.scalings_[:,i]**2).sum()**0.5 
                                for i in range(self._n_modes)]).T

        if not quiet:
            if self._n_modes > 1:
                LOGGER.debug('{0} modes were calculated in {1:.2f}s.'
                        .format(self._n_modes, time.time()-start))
            else:
                LOGGER.debug('{0} mode was calculated in {1:.2f}s.'
                        .format(self._n_modes, time.time()-start))

            if self._n_shuffles > 0:
                if self._n_modes > 1:
                    LOGGER.debug('Calculating {0} modes for {1} shuffles.'
                        .format(self._n_modes, self._n_shuffles))
                else:
                    LOGGER.debug('Calculating {0} mode for {1} shuffles.'
                        .format(self._n_modes, self._n_shuffles))
            
        self._shuffled_ldas = [LDA('shuffle '+str(n)) for n in range(self._n_shuffles)]
        self._coordsets_reshaped = self._coordsets.reshape(self._coordsets.shape[0], self._n_atoms, -1)

        n = 0
        while n < self._n_shuffles:
            labelsNew = self._labels.copy()
            # use random generator with None, 
            # then fresh, unpredictable entropy will be pulled from the OS
            rng = np.random.default_rng()
            rng.shuffle(labelsNew) # in place

            self._shuffled_ldas[n].calcModes(self._coordsets_reshaped, 
                                             labelsNew, quiet=True)
            
            if np.allclose(abs(np.dot(self._shuffled_ldas[n].getEigvecs()[0],
                                      self.getEigvecs()[0])), 
                           1):
                # LDA has flipped direction as labels match or are exactly flipped
                continue
            
            n += 1

        if self._n_shuffles > 0 and not quiet:
            if self._n_modes > 1:
                LOGGER.debug('{0} modes were calculated with {1} shuffles in {2:.2f}s.'
                        .format(self._n_modes, self._n_shuffles, time.time()-start))
            else:
                LOGGER.debug('{0} mode was calculated with {1} shuffles in {2:.2f}s.'
                        .format(self._n_modes, self._n_shuffles, time.time()-start))

    def addEigenpair(self, eigenvector, eigenvalue=None):
        """Add eigen *vector* and eigen *value* pair(s) to the instance.
        If eigen *value* is omitted, it will be set to 1.  Eigenvalues
        are set as variances."""

        NMA.addEigenpair(self, eigenvector, eigenvalue)
        self._vars = self._eigvals

    def setEigens(self, vectors, values=None):
        """Set eigen *vectors* and eigen *values*.  If eigen *values* are
        omitted, they will be set to 1.  Eigenvalues are set as variances."""

        self._clear()
        NMA.setEigens(self, vectors, values)
        self._vars = self._eigvals

    def getShuffledModes(self):
        return self._shuffled_ldas.copy()
    
    def getShuffledEigvecs(self):
        return np.array([lda.getEigvecs() for lda in self._shuffled_ldas])

    def getShuffledPercentile(self, percentile, take_abs=True):
        shuffles = self.getShuffledEigvecs()
        if take_abs:
            shuffles = abs(shuffles)
        return np.percentile(shuffles, percentile)



"""Extract a selection of atoms from a PDB file."""

from ..apptools import *

__all__ = ['prody_select']

def prody_select(selstr, *pdbs, **kwargs):
    """Write selected atoms from a PDB file in PDB format.

    :arg selstr: atom selection string, see :ref:`selections`

    :arg pdbs: PDB identifier(s) or filename(s)

    :arg output: output filename, default is :file:`pdb_selected.pdb`

    :arg prefix: prefix for output file, default is PDB filename

    :arg suffix: output filename suffix, default is :file:`_selected`"""

    from os.path import isfile
    from prody import LOGGER, parsePDB, writePDB

    #selstr = kwargs.get('selstr')
    if not pdbs:
        raise ValueError('pdb argument must be provided')

    if ((isfile(selstr) or len(selstr) == 4 and selstr[0].isdigit()) and
        len(pdbs) == 1 and not isfile(pdbs[0])):
        pdbs, selstr = selstr, pdbs[0]
        LOGGER.warn('The order of selstr and pdb arguments have switched '
                    'to support multiple files, old order will be supported '
                    'until v1.4.')
        pdbs = [pdbs]

    prefix = kwargs.get('prefix', None)
    suffix = kwargs.get('suffix', '_selected')
    output = kwargs.get('output', None)
    altloc = kwargs.get('altloc', 'all')
    uniteChains = kwargs.get('uniteChains', False)

    for pdb in pdbs:
        pdb = parsePDB(pdb, altloc=altloc, unite_chains=uniteChains)

        pdbselect = pdb.select(selstr)
        if pdbselect is None:
            LOGGER.warn('Selection {0} did not match any atoms.'
                        .format(repr(selstr)))
            return
        LOGGER.info('Selection {0} matched {1} atoms.'
                    .format(repr(selstr), len(pdbselect)))

        outname = output or ((prefix or pdb.getTitle()) + suffix)
        LOGGER.info('Selection is written into: ' +
                    writePDB(outname, pdbselect))



def addCommand(commands):

    subparser = commands.add_parser('select',
    help='select atoms and write a PDB file')

    subparser.add_argument('--quiet', help="suppress info messages to stderr",
        action=Quiet, nargs=0)

    subparser.add_argument('--examples', action=UsageExample, nargs=0,
        help='show usage examples and exit')

    subparser.set_defaults(usage_example=
    """This command selects specified atoms and writes them in a PDB file.

Fetch PDB files 1p38 and 1r39 and write backbone atoms in a file:

  $ prody select backbone 1p38 1r39""",
    test_examples=[0])


    group = subparser.add_argument_group('output options')

    group.add_argument('-o', '--output', dest='output', metavar='STR',
        type=str, help='output PDB filename (default: pdb_selected.pdb)')

    group.add_argument('-p', '--prefix', dest='prefix', metavar='STR',
        type=str, help=('output filename prefix (default: PDB filename)'))

    group.add_argument('-L', '--altloc', dest='altloc', metavar='STR',
        type=str, default='all', help=('altloc (default: %(default)s)'))

    group.add_argument('-x', '--suffix', dest='suffix', metavar='STR',
        type=str, default='_selected',
        help=('output filename suffix (default: %(default)s)'))

    group.add_argument('-u', '--unite-chains', dest='uniteChains',
        action='store_true',
        default=False, help=('unite chains if using mmCIF (default False)'))

    subparser.add_argument('select', help='atom selection string')
    subparser.add_argument('pdb', nargs='+',
        help='PDB identifier(s) or filename(s)')

    subparser.set_defaults(func=lambda ns: prody_select(ns.select,
                                                         *ns.pdb,
                                                         **ns.__dict__))
    subparser.set_defaults(subparser=subparser)



"""Refine MSA application."""

__author__ = 'Ahmet Bakan, Anindita Dutta'

from ..apptools import DevelApp
import numpy as np

__all__ = ['evol_rankorder']

APP = DevelApp('rankorder', 'identify highly coevolving pairs of residues')

APP.setExample(
"""This application identifies that top ranking pairs of residues that \
coevolve based on their mutual information.  By default coevolution is \
reported for pairs that are at least 3 residues apart in sequence. A z-score \
normalization can be applied to the mutinfo matrix to identify coevolving \
pairs.  The following examples show how to use with default as well as \
additional options:

    $ evol rankorder piwi_refined_mutinfo.txt -z

    $ evol rankorder piwi_refined_mutinfo.txt --msa piwi_refined.slx \
--label AGO6_ARATH""", [])


APP.addArgument('mutinfo',
    help='mutual information matrix')

APP.addGroup('input', 'input options')
APP.addArgument('-z', '--zscore',
    dest='zscore',
    action='store_true',
    help='apply zscore for identifying top ranked coevolving pairs',
    group='input'
    )
APP.addArgument('-d', '--delimiter',
    dest='delimiter',
    help='delimiter used in mutual information matrix file',
    type=str,
    metavar='STR',
    default=None,
    group='input'
    )
APP.addArgument('-p', '--pdb',
    dest='pdb',
    help='PDB file that contains same number of residues as the mutual '
    'information matrix, output residue numbers will be based on PDB file',
    default=None,
    type=str,
    metavar='STR',
    group='input'
    )
APP.addArgument('-m', '--msa',
    dest='msa',
    help='MSA file used for building the mutual info matrix, '
    'output residue numbers will be based on the most complete sequence '
    'in MSA if a PDB file or sequence label is not specified',
    default=None,
    type=str,
    metavar='STR',
    group='input'
    )
APP.addArgument('-l', '--label',
    dest='label',
    help='label in MSA file for output residue numbers',
    default=None,
    type=str,
    metavar='STR',
    group='input'
    )
APP.addGroup('output', 'output options')
APP.addArgument('-n', '--num-pairs',
    dest='numpairs',
    help='number of top ranking residue pairs to list',
    default=100,
    type=int,
    metavar='INT',
    group='output'
    )
APP.addArgument('-q', '--seq-sep',
    dest='seqsep',
    help='report coevolution for residue pairs that are sequentially '
    'separated by input value',
    default=3,
    type=int,
    metavar='INT',
    group='output'
    )
APP.addArgument('-t', '--min-dist',
    dest='dist',
    help='report coevolution for residue pairs whose CA atoms are spatially '
    'separated by at least the input value, used when a PDB file is given '
    'and --use-dist is true',
    default=10.0,
    type=float,
    metavar='FLOAT',
    group='output'
    )
APP.addArgument('-u', '--use-dist',
    dest='usedist',
    action='store_true',
    help='use structural separation to report coevolving pairs',
    group='output'
    )
APP.addArgument('-o', '--outname',
    dest='outname',
    help='output filename, default is mutinfo_rankorder.txt',
    type=str,
    metavar='STR',
    group='output'
    )

def calcAllDist(coordset):

    from prody import calcDistance
    shape = coordset.shape
    distance = np.zeros((shape[1], shape[1]))

    for i in range(shape[1]):
        temp = np.tile(coordset[0,i,:], (1, shape[1], 1))
        distance[:,i] = calcDistance(coordset, temp)

    return distance

def evol_rankorder(mutinfo, **kwargs):
    from prody import parseMSA, LOGGER, PY3K
    from prody import parsePDB, calcMSAOccupancy, trimAtomsUsingMSA
    from prody.utilities import openFile, splitSeqLabel
    from os.path import splitext

    delimiter = kwargs.get('delimiter')
    mi = np.loadtxt(str(mutinfo), delimiter=delimiter)

    ndim, shape = mi.ndim, mi.shape
    if ndim != 2 or shape[0] != shape[1]:
        raise ValueError('mutinfo must contain a square matrix')

    msa, label, msaflag = kwargs.get('msa'), kwargs.get('label'), False

    pdb, pdbflag = kwargs.get('pdb'), False

    resnum = None

    if msa is not None:
        msa = parseMSA(msa)
        if msa.numResidues() != shape[0]:
            LOGGER.info('Input MSA and mutinfo do not have similar no '
                        'of residues, ignoring MSA')
        else:
            index = msa.getIndex(label)
            try:
                if index is None:
                    if label is not None:
                        LOGGER.info('Could not find given label in MSA, '
                                    'using complete sequence from MSA')
                    occ = calcMSAOccupancy(msa._msa, 'row')
                    index = np.where(occ == occ.max())[0][0]
                    label, start, end = splitSeqLabel(msa[index].getLabel(True))
                else:
                    label, start, end = splitSeqLabel(msa[index].getLabel(True))
            except:
                LOGGER.info('Could not extract resnums from MSA')
            else:
                msaflag = True 

    if pdb is not None:
        from prody import parsePDB
        try:
            pdb = parsePDB(pdb)
        except:
            LOGGER.info('Could not parse PDB, ignoring PDB input')
        else:
            chains = list(pdb.iterChains())
            for chain in chains:
                sel = chain.select('protein and name CA')
                if sel.numAtoms() == shape[0]:
                    resnum = sel.getResnums()
                    coordset = sel.getCoordsets()
                    distance = calcAllDist(coordset)
                    pdbflag = True
                    label = pdb.getTitle()
                    LOGGER.info('Residue numbers will be based on pdb: '
                                '{0}'.format(pdb.getTitle()))
                    break
                else:
                    try:
                        sel = trimAtomsUsingMSA(sel, msa, chain=chain.getChid())
                        if sel.numAtoms() == shape[0]:
                            resnum = sel.getResnums()
                            coordset = sel.getCoordsets()
                            distance = calcAllDist(coordset)
                            pdbflag = True
                            label = pdb.getTitle()
                            LOGGER.info('Residue numbers will be based on pdb: '
                                        '{0}'.format(pdb.getTitle()))
                            break
                    except:
                        LOGGER.info('Number of residues in PDB does not match '
                                    'mutinfo matrix and no MSA was provided to '
                                    'align the PDB against, so ignoring PDB input')

    if not pdbflag:
        if msaflag:
            if (start and end is not None) and (start < end):
                resnum = np.arange(start, end+1)
                if len(resnum) != shape[0]:
                    LOGGER.info('Label: {0}/{1}-{2} and mutinfo do '
                                'not have similar no of residues, using '
                                'serial indexing'.format(label, start, end))
                    label = 'Serial Index'
                    resnum = np.arange(1, shape[0]+1)
                else:
                    LOGGER.info('Residue numbers will be based on MSA and label: '
                                '{0}'.format(label))
            else:
                LOGGER.info('Could not identify residue indexes from MSA'
                                ' using serial indexing')
                label = 'Serial Index'
                resnum = np.arange(1, shape[0]+1)
        else:
            LOGGER.info('MSA or PDB not given or does not match mutinfo, '
                        'using serial indexing')
            resnum = np.arange(1, shape[0]+1)

    LOGGER.info('Residue numbers start and end with {0}-{1}'.
                format(str(resnum[0]), str(resnum[-1])))

    outname = kwargs.get('outname')
    if outname is None:
        outname, ext = splitext(str(mutinfo))
        if ext.lower() == '.gz':
            outname, _ = splitext(str(mutinfo))
    else:
        outname, ext = splitext(str(outname))
        if ext is None:
            ext = '.txt'

    outname += '_rankorder' + ext
    zscore = kwargs.get('zscore')
    if zscore:
        LOGGER.info('zscore normalization applied such that each column '
                    'has 0 mean and standard deviation 1')
        header = 'Serial\tRow\tColumn\tZscore'
        mi = (mi - mi.mean(0)) / mi.std(0)
    else:
        header = 'Serial\tRow\tColumn\tMI'

    mi_ind_start, mi_ind_end = np.tril_indices(shape[0], k=-1)
    mi_matrix = mi[mi_ind_start, mi_ind_end]
    sorted_index = mi_matrix.argsort(axis=None)[::-1]
    row = mi_ind_start[sorted_index]
    column = mi_ind_end[sorted_index]
    count = 1
    i = 0

    if PY3K:
        mode = 'w'
    else:
        mode = 'wb'
    f = openFile(outname, mode)
    
    if label is None:
        label = 'Serial Index'

    numpairs = kwargs.get('numpairs')
    size = len(row)
    seqsep = kwargs.get('seqsep')
    if not kwargs.get('usedist') or not pdbflag:
        if kwargs.get('usedist'):
            LOGGER.info('use-struct-sep set to true, but PDB not given or '
                        'incorrect residue number. Using sequence separation')
        else:
            if pdbflag:
                LOGGER.info('use-dist not set, using sequence separation'
                            ' to report coevolving pairs')
        f.write(('Label: '+ label + '\t' + 'Residue Numbers: ' +
             str(resnum[0]) + '-' + str(resnum[-1]) + '\tSequence Separation:' +
             str(seqsep) + '\n'))
        if pdbflag:
            f.write((header + '\tDistance\n'))
            while count <=numpairs  and i < size:
                if row[i] > (column[i] + seqsep):
                    f.write('{0}\t{1}\t{2}\t{3:.3f}\t{4:.2f}\n'.
                            format(count, resnum[row[i]], resnum[column[i]],
                                   mi[row[i], column[i]],
                                   distance[row[i], column[i]]))
                    count += 1
                i += 1
        else:
            f.write((header + '\n'))
            while count <=numpairs  and i < size:
                if row[i] > (column[i] + seqsep):
                    f.write('{0}\t{1}\t{2}\t{3:.3f}\n'.
                            format(count, resnum[row[i]], resnum[column[i]],
                                   mi[row[i], column[i]]))
                    count += 1
                i += 1
    else:
        structsep = kwargs.get('dist')
        f.write(('Label: '+ label + '\t' + 'Residue Numbers: ' +
             str(resnum[0]) + '-' + str(resnum[-1]) + 'Distance Cutoff:' +
             str(structsep) + '\n'))
        f.write((header + '\tDistance\n'))
        while count <=numpairs  and i < size:
            if distance[row[i], column[i]] > structsep:
                f.write('{0}\t{1}\t{2}\t{3:.3f}\t{4:.2f}\n'.
                        format(count, resnum[row[i]], resnum[column[i]],
                               mi[row[i], column[i]],
                               distance[row[i], column[i]]))
                count += 1
            i += 1
    f.close()

APP.setFunction(evol_rankorder)



# -*- coding: utf-8 -*-
"""This module defines classes and functions to fetch, parse, and write
structural data files, execute structural analysis programs, and to access
and search structural databases, e.g. `ProteinDataBank <http://wwpdb.org>`_.

PDB resources
=============

  * :func:`.fetchPDB` - retrieve PDB files
  * :func:`.fetchPDBviaFTP` - download PDB/PDBML/mmCIF files
  * :func:`.fetchPDBviaHTTP` - download PDB files

You can use following functions to manage PDB file resources:

  * :func:`.pathPDBFolder` - local folder for storing PDB files
  * :func:`.pathPDBMirror` - local PDB mirror path
  * :func:`.wwPDBServer` - set wwPDB FTP/HTTP server for downloads

The following functions can be used to handle local PDB files:

  * :func:`.findPDBFiles` - return a dictionary containing files in a path
  * :func:`.iterPDBFilenames` - yield file names in a path or local PDB mirror


Blast search PDB
================

The following are for blast searching PDB content.

  * :func:`.blastPDB` - blast search NCBI PDB database
  * :class:`.PDBBlastRecord` - store/evaluate NCBI PDB blast search results

PDB clusters biopolymer chains using blast weekly.  These clusters can be
retrieved using the following functions.  Using cluster data is as good
as blast searching PDB most of the time and incredibly faster always.

  * :func:`.listPDBCluster` - get list of identifiers in a PDB sequence cluster
  * :func:`.loadPDBClusters` - load PDB clusters into memory
  * :func:`.fetchPDBClusters` - retrieve PDB sequence cluster data from wwPDB

Parse/write PDB files
=====================

Following ProDy functions are for parsing and writing :file:`.pdb` files:

  * :func:`.parsePDB` - parse :file:`.pdb` formated file
  * :func:`.parsePDBStream` - parse :file:`.pdb` formated stream
  * :func:`.writePDB` - write :file:`.pdb` formatted file
  * :func:`.writePDBStream`  write :file:`.pdb` formated stream

Since :file:`.pqr` format is similar to :file:`.pdb` format, following
functions come as bonus features:

  * :func:`.writePQR` - write atomic data to a file in :file:`.pqr` format
  * :func:`.parsePQR` - parse atomic data from files in :file:`.pqr` format


.. seealso::

   Atom data (coordinates, atom names, residue names, etc.) parsed from
   PDB/PSF/PQR/mmCIF files are stored in :class:`~.AtomGroup` instances.
   See :mod:`~prody.atomic` module documentation for more details.

Parse mmCIF files
=====================

Following ProDy functions are for parsing :file:`.cif` files:

  * :func:`.parseMMCIF` - parse :file:`.cif` formated file
  * :func:`.parseMMCIFStream` - parse :file:`.cif` formated stream

.. seealso::

   Atom data (coordinates, atom names, residue names, etc.) parsed from
   PDB/PSF/PQR/mmCIF files are stored in :class:`~.AtomGroup` instances.
   See :mod:`~prody.atomic` module documentation for more details.

Quick visualization
===================

:func:`.showProtein` function can be used to take a quick look at protein
structures.

Edit structures
===============

The following functions allow editing structures using structural data from PDB
header records:

  * :func:`.assignSecstr` - add secondary structure data from header to atoms
  * :func:`.buildBiomolecules` - build biomolecule from header records


PDB header data
===============

Use the following to parse and access header data in PDB files:

  * :func:`.parsePDBHeader` - parse header data from :file:`.pdb` files
  * :class:`.Chemical` - store PDB chemical (heterogen) component data
  * :class:`.Polymer` - store PDB polymer (macromolecule) component data
  * :class:`.DBRef` - store polymer sequence database reference records

Analyze interactions and stability with InSty and find water bridges with WatFinder
====================

Use the following to analyze interactions within protein structure or
between protein and ligand structure in single PDB file or in trajectory:

  * :func:`.addHydrogens` - add missing hydrogens to :file:`.pdb` files
  * :func:`.calcHydrogenBonds` - compute hydrogen bonds in proteins
  * :func:`.calcSaltBridges` - compute salt bridges in proteins
  * :func:`.calcRepulsiveIonicBonding` - compute repulsive ionic bonding in proteins
  * :func:`.calcPiStacking` - compute Pi-stacking interactions in proteins
  * :func:`.calcPiCation` - compute Pi-cation interactions in proteins
  * :func:`.calcHydrophobic` - compute hydrophobic interactions in proteins
  * :func:`.calcProteinInteractions` - compute all above interaction types at once
  * :func:`.showProteinInteractions_VMD` - return TCL file for visualization in VMD 

  * :func:`.calcHydrogenBondsDCD` - compute hydrogen bonds in a trajectory for proteins
  * :func:`.calcSaltBridgesDCD` - ompute salt bridges in a trajectory for proteins
  * :func:`.calcRepulsiveIonicBondingDCD` - compute repulsive ionic bonding in a trajectory for proteins    
  * :func:`.calcPiStackingDCD` - compute Pi-stacking interactions in a trajectory for proteins
  * :func:`.calcPiCationDCD` - compute Pi-cation interactions in a trajectory for proteins
  * :func:`.calcHydrophobicDCD` - compute hydrophobic interactions in a trajectory for proteins
  * :func:`.calcStatisticsInteractions` - return statistical information for each interaction type 

  * :func:`.calcLigandInteractions` - compute all type of interactions between protein and ligand
  * :func:`.listLigandInteractions` - return list of interactions between protein and ligand
  * :func:`.showLigandInteraction_VMD` - return TCL file for visualization of interactions for VMD

  * :class:`.Interactions` - store inteactions for a single PDB structure
  * :class:`.InteractionsDCD` - store interactions for a trajectory

Compare/align chains
====================

The following functions can be used to match, align, and map polypeptide chains:

  * :func:`.matchChains` - finds matching chains in two protein structures
  * :func:`.matchAlign` - finds best matching chains and aligns structures
  * :func:`.mapOntoChain` - maps chains in a structure onto a reference chain

The following functions can be used to adjust alignment parameters:

  * :func:`.getAlignmentMethod`, :func:`.setAlignmentMethod`
  * :func:`.getMatchScore`, :func:`.setMatchScore`
  * :func:`.getMismatchScore`, :func:`.setMismatchScore`
  * :func:`.getGapPenalty`, :func:`.setGapPenalty`
  * :func:`.getGapExtPenalty`, :func:`.setGapExtPenalty`

Execute DSSP
============

The following functions can be used to execute DSSP structural analysis program
and/or parse results:

  * :func:`.execDSSP` - execute :program:`dssp`
  * :func:`.performDSSP` - execute :program:`dssp` and parse results
  * :func:`.parseDSSP` - parse structural data from :program:`dssp` output

Execute STRIDE
==============

The following functions can be used to execute STRIDE structural analysis program
and/or parse results:

  * :func:`.execSTRIDE` - execute :program:`stride`
  * :func:`.parseSTRIDE` - parse structural data from :program:`stride` output
  * :func:`.performSTRIDE` - execute :program:`stride` and parse results

Handle EMD Map Files and Build Pseudoatoms into them
===========

Use the following to parse and access header data in EMD files:

  * :func:`.parseEMD` - parse structural data from :file:`.emd` files
  * :class:`.EMDMAP` - access structural data from :file:`.emd` files
  * :class:`.TRNET` - fit pseudoatoms to EM density maps using the TRN algorithm

Add missing atoms including hydrogens
===========

Use the following to add missing atoms

  * :func:`.addMissingAtoms` - add missing atoms with separately installed OpenBabel or PDBFixer

Missing residues can also be added if a PDB or mmCIF file with SEQRES entries is provided.
"""

__all__ = []

from . import compare
from .compare import *
__all__.extend(compare.__all__)

from . import localpdb
from .localpdb import *
__all__.extend(localpdb.__all__)

from . import wwpdb
from .wwpdb import *
__all__.extend(wwpdb.__all__)

from . import pdbclusters
from .pdbclusters import *
__all__.extend(pdbclusters.__all__)

from . import blastpdb
from .blastpdb import *
__all__.extend(blastpdb.__all__)

from . import functions
from .functions import *
__all__.extend(functions.__all__)

from . import header
from .header import *
__all__.extend(header.__all__)

from . import cifheader
from .cifheader import *
__all__.extend(cifheader.__all__)

from . import dssp
from .dssp import *
__all__.extend(dssp.__all__)

from . import stride
from .stride import *
__all__.extend(stride.__all__)

from . import pdbfile
from .pdbfile import *
__all__.extend(pdbfile.__all__)

from . import mmtffile
from .mmtffile import *
__all__.extend(mmtffile.__all__)

from . import emdfile
from .emdfile import *
__all__.extend(emdfile.__all__)

from . import ciffile
from .ciffile import *
__all__.extend(ciffile.__all__)

_getPDBid  = ciffile._getPDBid = pdbfile._getPDBid
long_id_check_str = ciffile.long_id_check_str = pdbfile.long_id_check_str
wwpdb.long_id_check_str = long_id_check_str

from . import starfile
from .starfile import *
__all__.extend(starfile.__all__)

from . import interactions
from .interactions import *
__all__.extend(interactions.__all__)

try:
    from . import waterbridges
    from .waterbridges import *
except SyntaxError:
    import logging
    logger = logging.getLogger()
    logger.warn("Cannot import waterbridges")
else:
    __all__.extend(waterbridges.__all__)

from . import fixer
from .fixer import *
__all__.extend(fixer.__all__)

from .pdbfile import PDBParseError
from .ciffile import MMCIFParseError

from . import opm
from .opm import *
__all__.extend(opm.__all__)



# -*- coding: utf-8 -*-
"""This module defines classes to handle individual atoms."""

import numpy as np

from .flags import PLANTERS
from .fields import ATOMIC_FIELDS, READONLY
from .fields import wrapGetMethod, wrapSetMethod
from .pointer import AtomPointer
from .bond import Bond

__all__ = ['Atom']


class Atom(AtomPointer):

    """A class for handling individual atoms in an :class:`.AtomGroup`."""

    __slots__ = ['_ag', '_acsi', '_index']

    def __init__(self, ag, index, acsi):
        AtomPointer.__init__(self, ag, acsi)
        self._index = int(index)

    def __repr__(self):

        n_csets = self._ag.numCoordsets()
        if n_csets == 1:
            return '<Atom: {0} from {1} (index {2})>'.format(
                   self.getName(), self._ag.getTitle(), self._index)
        elif n_csets > 1:
            return ('<Atom: {0} from {1} (index {2}; active #{3} of '
                    '{4} coordsets)>').format(self.getName(),
                     self._ag.getTitle(), self._index, self.getACSIndex(),
                     n_csets)
        else:
            return ('<Atom: {0} from {1} (index {2}; no coordinates)>'
                    ).format(self.getName(), self._ag.getTitle(), self._index)

    def __str__(self):

        return 'Atom {0} (index {1})'.format(self.getName(), self._index)

    def __len__(self):

        return 1

    def __int__(self):

        return self._index

    def numAtoms(self, flag=None):
        """Returns number of atoms, or number of atoms with given *flag*."""

        return len(self._getSubset(flag)) if flag else 1

    def getIndex(self):
        """Returns index of the atom."""

        return self._index

    def getIndices(self):
        """Returns index of the atom in an :class:`numpy.ndarray`."""

        return np.array([self._index])

    _getIndices = getIndices

    def iterAtoms(self):
        """Yield atoms."""

        yield Atom(ag=self._ag, index=self._index, acsi=self.getACSIndex())

    __iter__ = iterAtoms

    def getCoords(self):
        """Returns a copy of coordinates of the atom from the active coordinate
        set."""

        if self._ag._coords is not None:
            return self._ag._coords[self.getACSIndex(), self._index].copy()

    def _getCoords(self):
        """Returns a view of coordinates of the atom from the active coordinate
        set."""

        if self._ag._coords is not None:
            return self._ag._coords[self.getACSIndex(), self._index]

    def setCoords(self, coords):
        """Set coordinates of the atom in the active coordinate set."""

        acsi = self.getACSIndex()
        self._ag._coords[acsi, self._index] = coords
        self._ag._setTimeStamp(acsi)

    def getAnisou(self):
        """Returns a copy of anisotropic temperature factors of the atom from the active coordinate
        set."""

        if self._ag._anisous is not None:
            return self._ag._anisous[self.getACSIndex(), self._index].copy()

    def _getAnisou(self):
        """Returns a view of anisotropic temperature factors of the atom from the active coordinate
        set."""

        if self._ag._anisous is not None:
            return self._ag._anisous[self.getACSIndex(), self._index]

    def setAnisou(self, anisou):
        """Set anisotropic temperature factors of the atom in the active coordinate set."""

        acsi = self.getACSIndex()
        self._ag._anisous[acsi, self._index] = anisou
        self._ag._setTimeStamp(acsi)

    def getCoordsets(self, indices=None):
        """Returns a copy of coordinate set(s) at given *indices*."""

        if self._ag._coords is None:
            return None

        if indices is None:
            return self._ag._coords[:, self._index].copy()

        if isinstance(indices, (int, slice)):
            return self._ag._coords[indices, self._index].copy()

        if isinstance(indices, (list, np.ndarray)):
            return self._ag._coords[indices, self._index]

        raise IndexError('indices must be an integer, a list/array of '
                         'integers, a slice, or None')

    def _getCoordsets(self, indices=None):
        """Returns a view of coordinate set(s) at given *indices*."""

        if self._ag._coords is None:
            return None

        if indices is None:
            indices = slice(None)

        return self._ag._coords[indices, self._index]

    def iterCoordsets(self):
        """Yield copies of coordinate sets."""

        for i in range(self.numCoordsets()):
            yield self._ag._coords[i, self._index].copy()

    def _iterCoordsets(self):
        """Yield views of coordinate sets."""

        for i in range(self.numCoordsets()):
            yield self._ag._coords[i, self._index]

    def getMasses(self):
        """get the mass atom. """
        from prody.utilities.misctools import getMasses
        
        return getMasses(self.getElement())

    def getData(self, label):
        """Returns a copy of data associated with *label*, if it is present."""

        try:
            data = self._ag._getData(label)
        except KeyError:
            pass
        else:
            if data.ndim > 1:
                return data[self._index]
            else:
                return data[self._index].copy()

    _getData = getData

    def setData(self, label, data):
        """Update *data* associated with *label*.

        :raise AttributeError: when *label* is not in use or read-only"""

        if label in READONLY:
            raise AttributeError('{0} is read-only'.format(repr(label)))
        if label in ATOMIC_FIELDS:
            getattr(self, 'set' + ATOMIC_FIELDS[label].meth)(data)
        else:
            try:
                self._ag._data[label][self._index] = data
            except KeyError:
                raise AttributeError('data with label {0} must be set for'
                                       ' AtomGroup first'.format(repr(label)))

    def getFlag(self, label):
        """Returns atom flag."""

        return self._ag._getFlags(label)[self._index]

    def setFlag(self, label, value):
        """Update flag associated with *label*.

         :raise AttributeError: when *label* is not in use or read-only"""

        if label in PLANTERS:
            raise AttributeError('flag {0} cannot be changed by user'
                                    .format(repr(label)))
        flags = self._ag._getFlags(label)
        if flags is None:
            raise AttributeError('flags with label {0} must be set for '
                                    'AtomGroup first'.format(repr(label)))
        flags[self._index] = value

    def getSelstr(self):
        """Returns selection string that will select this atom."""

        return 'index {0}'.format(self._index)

    def numBonds(self):
        """Returns number of bonds formed by this atom.  Bonds must be set first
        using :meth:`.AtomGroup.setBonds`."""

        numbonds = self._ag._data.get('numbonds')
        if numbonds is not None:
            return numbonds[self._index]

    def iterBonds(self):
        """Yield bonds formed by the atom.  Use :meth:`setBonds` for setting
        bonds."""

        ag = self._ag
        acsi = self.getACSIndex()
        for bond in self._iterBonds():
            yield Bond(ag, bond, acsi)

    def _iterBonds(self):
        """Yield pairs of bonded atom indices."""

        ag = self._ag
        if ag._bmap is None:
            raise ValueError('bonds are not set, use `AtomGroup.setBonds`')

        this = self._index
        for other in ag._bmap[this]:
            if other == -1:
                break
            yield this, other

    def iterBonded(self):
        """Yield bonded atoms.  Use :meth:`setBonds` for setting bonds."""

        ag = self._ag
        if ag._bmap is None:
            raise ValueError('bonds are not set, use `AtomGroup.setBonds`')

        acsi = self.getACSIndex()
        this = self._index
        for other in self._ag._bmap[this]:
            if other == -1:
                break
            yield Atom(ag, other, acsi)

    def toTEMPyAtom(self):
        """Returns a TEMPy BioPyAtom or Structure object as appropriate"""  
        try:
            from TEMPy.protein.prot_rep_biopy import Atom as TEMPyAtom
        except ImportError:
            raise ImportError('TEMPy is needed for this functionality')

        return TEMPyAtom(
            self.getName(), self.getCoords(),
            'HETATM' if self.getFlag('hetatm') else 'ATOM',
            self.getSerial(), self.getBeta(),
            self.getAltloc(), self.getIcode(),
            self.getCharge(), self.getElement(),
            self.getOccupancy(), self.getResname(),
            self.getACSIndex(), self.getChid(),
            self.getResnum())


for fname, field in ATOMIC_FIELDS.items():

    if field.private:
        continue

    meth = field.meth
    getMeth = 'get' + meth
    setMeth = 'set' + meth
    # Define public method for retrieving a copy of data array
    def getData(self, meth=field.meth_pl, call=field.call):
        data = getattr(self._ag, '_get' + meth)()
        if data is not None:
            return data[self._index]
    getData = wrapGetMethod(getData)
    getData.__name__ = getMeth
    getData.__doc__ = field.getDocstr('get', False)
    setattr(Atom, getMeth, getData)
    setattr(Atom, '_' + getMeth, getData)

    if field.readonly:
        continue

    # Define public method for setting values in data array
    def setData(self, value, var=fname, none=field.none):
        array = self._ag._data[var]
        if array is None:
            raise AttributeError('attribute of the AtomGroup is '
                                 'not set')
        array[self._index] = value
        if none: self._ag._none(none)
    setData = wrapSetMethod(setData)
    setData.__name__ = setMeth
    setData.__doc__ = field.getDocstr('set', False)
    setattr(Atom, setMeth, setData)

del getData
del setData
